# cython: wraparound=False
# cython: boundscheck=False
# cython: nonecheck=False

# Authors: Thierry Moudiki
#
# License: BSD 3

import functools
import numpy as np
cimport numpy as np
cimport cython
import gc
from copy import deepcopy
from tqdm import tqdm

from libc.math cimport log, exp, sqrt
from numpy.linalg import lstsq
from scipy.linalg import solve


# 0 - utils -----

# 0 - 0 data structures & funcs -----

ctypedef fused nparray_double:    
    double


cdef public double call_f(f, x):
    f(x)

    
cdef public double[:] calc_grad(f, x):
    return numerical_gradient(f, x)

    
cdef public double[:, :] calc_hessian(f, x):
    return numerical_hessian(f, x)   

DTYPE_double = np.double

DTYPE_int = np.int

ctypedef np.double_t DTYPE_double_t

def backtrack_line_search(object f, double[:] x, double[:] p_k, double[:] grad_k, 
                          double rho = 0.5, double c = 1e-4, **kwargs):
    
    cdef int i = 0
    cdef double alpha = 1.0  
    cdef int num_iters = 100
    
    f_ = lambda x: f(x, **kwargs)    
    
    for i in range(num_iters):        
        if (call_f(f_, calc_fplus(x, alpha, p_k)) <= (call_f(f_, x) + c*alpha*crossprod(grad_k, p_k))):
            break
        alpha = rho*alpha        
        
    return alpha


cdef public double calc_learning_rate(object f, double[:] x, 
                                      double[:] p_k, double[:] grad_k):
    return backtrack_line_search(f, x, p_k, grad_k)
    

def generate_index(double[:] response, double batch_prop=1.0, 
                   randomization="strat", seed=123):
    """Generation of indices for Stochastic gradient descent."""
    
    cdef long int n = len(response)
    cdef long int size_out = 0
    
    if batch_prop < 1:
    
        if randomization=="strat":        
            return subsample2(response, batch_prop, seed=seed)
        
        if randomization=="shuffle":                    
            np.random.seed(seed)        
            return np.asarray(np.floor(np.random.choice(range(n), 
                                    size=np.floor(batch_prop*n), 
                                    replace=False)), dtype=np.int)
    
    return None


cdef int check_is_min(f, double[:] x):
    # print("hessian eigen values > 0")
    # print(numerical_hessian(f, x)[1])
    return (numerical_hessian(f, x)[1])*1

cdef list flatten(x): 
            
    cdef list res = []    
    
    for elt1 in x:
                    
        try:
            
            for elt2 in elt1:
                
                res.append(elt2)  
                
        except:
            
            res.append(elt1)  
                                
    return res


def is_factor(y):

    cdef long int n = len(y)
    cdef long int idx = 0
    ans = True    
    
    # check if x is int
    def is_int(x):
        try:
            return int(x) == x
        except:
            return False

    # check if x is float
    def is_float(x):
        return isinstance(x, float)
    
    for idx in range(n):
        
        if is_int(y[idx]) & (is_float(y[idx]) == False):
            
            idx += 1
            
        else:
            
            ans = False
            
            break

    return ans

cdef double crossprod(double[:] x, double[:] y):
    
    cdef double res 
    cdef long int i, n 
    
    n = len(y)
    res = 0 
    
    for i in range(len(x)):
        res += x[i]*y[i]
    
    return res 


cdef double[:] calc_fplus(double[:] x, double alpha, double[:] p_k):        
    
    cdef long int p = len(x)
    cdef double[:] res = np.zeros(p)     
    cdef long int j = 0
    
    for j in range(p):
        res[j] = x[j] + alpha*p_k[j]
        
    return res 

# Useful functions -----

# 1 - gradient-----
        
def numerical_gradient(object f, nparray_double[:] x, **kwargs):
    
    cdef long int p = len(x)
    cdef long int ix = 0    
    cdef double eps_factor = 6.055454452393343e-06 # machine eps ** (1/3)
    cdef double zero = 2.220446049250313e-16
    cdef double value_x, h, fx_plus, fx_minus
    cdef nparray_double[:] res = np.zeros_like(x)
    
    f_ = lambda x: f(x, **kwargs)
       
    for ix in range(p):
        
        value_x = x[ix]       
        
        h = max(eps_factor*value_x, 1e-8)
        
        x[ix] = value_x + h
        #fx_plus = f(x, **kwargs)
        fx_plus = call_f(f_, x)
        
        x[ix] = value_x - h
        #fx_minus = f(x, **kwargs)
        fx_minus = call_f(f_, x)
        
        x[ix] = value_x # restore
        
        res[ix] = (fx_plus - fx_minus)/(2*h)        
                    
    return np.asarray(res)



# 2 - hessian-----
    
def numerical_hessian(object f, nparray_double[:] x, **kwargs):
    
    cdef long int p = len(x)
    cdef long int ix, jx = 0    
    cdef double eps_factor = 0.0001220703125 # machine eps ** (1/4)
    cdef double zero = 2.220446049250313e-16
    cdef double value_x, value_y, h, k
    cdef double fx, fx_plus, fx_minus, fx_plus_minus, fx_minus_plus
    cdef double[:,::1] H = np.zeros((p, p))
    cdef double temp = 0 
    
    
    f_ = lambda x: f(x, **kwargs)
    
    fx = call_f(f_, x)
       
    for ix in range(p):
        
        for jx in range(ix, p):
            
            if (ix < jx):
                
                value_x = x[ix]            
                value_y = x[jx]
                
                h = max(eps_factor*value_x, 1e-8)
                k = max(eps_factor*value_y, 1e-8) 
                
                x[ix] = value_x + h
                x[jx] = value_y + k
                fx_plus = call_f(f_, x)
                
                x[ix] = value_x + h
                x[jx] = value_y - k
                fx_plus_minus = call_f(f_, x)
                
                x[ix] = value_x - h
                x[jx] = value_y + k
                fx_minus_plus = call_f(f_, x)
                
                x[ix] = value_x - h
                x[jx] = value_y - k
                fx_minus = call_f(f_, x)
                
                x[ix] = value_x # restore
                x[jx] = value_y # restore
                
                temp = (fx_plus - fx_plus_minus - fx_minus_plus + fx_minus)/(4*h*k)        
                                
            
            else:
                
                value_x = x[ix]                                                            
                
                h = max(eps_factor*value_x, 1e-8)
                
                x[ix] = value_x + h
                fx_plus = call_f(f_, x)
                                
                x[ix] = value_x - h
                fx_minus = call_f(f_, x)
                
                x[ix] = value_x # restore
                
                temp = (fx_plus - 2*fx + fx_minus)/(h**2)        
    
            
            H[ix, jx] = temp            
            H[jx, ix] = temp                
                
    res = np.asarray(H)
                    
    return res, all(np.linalg.eig(res)[0] > 0)


# 3 - stratified subsampling based on the response -----

def subsample2(y, double row_sample=0.8, int seed=123):
    
    cdef long int i, n_classes, n_obs, n_obs_out
    cdef long int[:] n_elem_classes
    cdef double[:] freqs_hist
    cdef list index = []

    assert (row_sample < 1) & (
        row_sample >= 0
    ), "'row_sample' must be < 1 and >= 0"

    n_obs = len(y)
    n_obs_out = np.ceil(n_obs * row_sample)

    # preproc -----
    if is_factor(y):

        classes, n_elem_classes = np.unique(y, return_counts=True)
        n_classes = len(classes)
        y_as_classes = y.copy()
        freqs_hist = np.zeros_like(n_elem_classes, dtype=np.float)
        
        for i in range(len(n_elem_classes)):
            freqs_hist[i] = np.float(n_elem_classes[i]) / n_obs

    else:

        h = np.histogram(y, bins="auto")
        n_elem_classes = np.asarray(h[0], dtype=np.int)
        freqs_hist = np.zeros_like(n_elem_classes, dtype=np.float)
        
        for i in range(len(n_elem_classes)):
            freqs_hist[i] = np.float(n_elem_classes[i]) / n_obs
        
        breaks = h[1]

        n_breaks_1 = len(breaks) - 1
        classes = range(n_breaks_1)
        n_classes = n_breaks_1
        y_as_classes = np.zeros_like(y, dtype=int)

        for i in classes:
            y_as_classes[(y > breaks[i]) * (y <= breaks[i + 1])] = int(i)

    # main loop ----

    np.random.seed(seed)

    for i in range(n_classes):

        bool_class_i = (y_as_classes == classes[i])               

        # index_class_i = [i for i, e in enumerate(bool_class_i) if e == True]
        index_class_i = np.asarray(np.where(bool_class_i == True)[0], dtype=np.int)

        if np.sum(bool_class_i) > 1:  # at least 2 elements in class  #i

            index.append(
                np.random.choice(
                    index_class_i,
                    size=int(n_obs_out * freqs_hist[i]),  # output size
                    replace=True,
                ).tolist()
            )

        else:  # only one element in class

            try:

                index.append(index_class_i[0])

            except:

                0
    
    return np.asarray(flatten(index))


# 4 - One-hot encoder -----

def one_hot_encode(long int[:] y, 
                   int n_classes):
    
    cdef long int i 
    cdef long int n_obs = len(y)
    cdef double[:,::1] res = np.zeros((n_obs, n_classes), dtype=DTYPE_double)        

    for i in range(n_obs):
        res[i, y[i]] = 1

    return np.asarray(res)


# Coordinate descent (Stochastic) -----

# 1 - algos -----

def scd(loss_func, double[:] response, double[:] x, int num_iters=200, 
        double batch_prop=1.0, double learning_rate=0.01, double mass=0.9, 
        double decay=0.1, method="momentum", randomization="strat", 
        double tolerance=1e-3, verbose=1, 
        **kwargs):
    """Stochastic gradient descent with momentum and adaptive learning rates."""
    
    cdef int i = 0 
    cdef long int j = 0
    cdef long int n = len(response)
    cdef long int p = len(x)
    cdef double[:] velocity = np.zeros(p)
    cdef double grad_x, decay_rate, learning_rate_, h0
    cdef list losses = []

    if verbose == 1:
        iterator = tqdm(range(num_iters))          
    else:
        iterator = range(num_iters)
    
    f = lambda x: loss_func(x, **kwargs)              
       
    if method is "momentum":                

        for i in iterator: 
                   
            idx = generate_index(response=response, batch_prop=batch_prop, 
                                 randomization=randomization, 
                                 seed=i)   

            def f_j(double h, double[:] xx, long int j):        
                cdef double value_x = 0        
                cdef double res = 0 
                value_x = xx[j]            
                xx[j] = xx[j] + h
                res = loss_func(xx, row_index=idx, **kwargs)          
                xx[j] = value_x
                return res                                                                                     
            
            diff = -np.asarray(x)

            if verbose == 2:
                print(f"\n x prev: {np.asarray(x)}")
            
            for j in range(p):                          
                h0 = 6.055454452393343e-06*x[j]        
                grad_x = (f_j(h0, x, j) - f_j(-h0, x, j))/(2*h0)        
                velocity[j] = mass * velocity[j] - learning_rate * grad_x  
                x[j] = x[j] + velocity[j]        
            
            diff += np.asarray(x)  

            if verbose == 2:
                print(f"\n x new: {np.asarray(x)}")
            
            losses.append(f(x))

            if (len(losses) > 3) and (np.abs(np.diff(losses[-2:])[0]) < tolerance):
                break
            
            if verbose == 2:

                print("\n")
                print(f"iter {i+1} - decrease -----")

                try: 
                    print(np.linalg.norm(diff, 1)) 
                except:
                    pass   

                print(f"iter {i+1} - loss -----")
                print(np.flip(losses)[0])     
    
    if method in ("exp", "poly"):
        
        for i in iterator:
            
            idx = generate_index(response=response, batch_prop=batch_prop, 
                                 randomization=randomization, 
                                 seed=i)
            
            def f_j(double h, double[:] xx, long int j):        
                cdef double value_x = 0        
                cdef double res = 0 
                value_x = xx[j]            
                xx[j] = xx[j] + h
                res = loss_func(xx, row_index=idx, **kwargs)          
                xx[j] = value_x
                return res      
                                                
            decay_rate = (1 + decay*i) if method is "poly" else exp(decay*i)
            
            diff = -np.asarray(x)

            if verbose == 2:
                print(f"\n x prev: {np.asarray(x)}")
            
            losses.append(f(x))
            
            for j in range(p):  
                h0 = 6.055454452393343e-06*x[j]        
                grad_x = (f_j(h0, x, j) - f_j(-h0, x, j))/(2*h0)        
                x[j] = x[j] - grad_x*learning_rate/decay_rate
                
            diff += np.asarray(x) 

            if verbose == 2:
                print(f"\n x new: {np.asarray(x)}") 
            
            if (len(losses) > 3) and (np.abs(np.diff(losses[-2:])[0]) < tolerance):
                break

            if verbose == 2:

                print("\n")
                print(f"iter {i+1} - decrease -----")

                try:
                    print(np.linalg.norm(diff, 1))   
                except:
                    pass 

                print(f"iter {i+1} - loss -----")
                print(np.flip(losses)[0])

                    
    return np.asarray(x), num_iters, losses


# Gradient descent (Stochastic) -----  

def sgd(loss_func, double[:] response, double[:] x, int num_iters=200, 
        double batch_prop=1.0, double learning_rate=0.01, double mass=0.9, 
        double decay=0.1, method="momentum", randomization="strat", 
        double tolerance=1e-3, verbose=1, 
        **kwargs):
    """Stochastic gradient descent with momentum and adaptive learning rates."""
    
    cdef int i = 0 
    cdef long int j = 0
    cdef long int n = len(response)
    cdef long int p = len(x)
    cdef double[:] velocity = np.zeros(p)
    cdef double[:] grad_i = np.zeros(p)
    cdef double decay_rate, learning_rate_
    cdef list losses = []
    
    if verbose == 1:
        iterator = tqdm(range(num_iters))          
    else:
        iterator = range(num_iters)
            
    f = lambda x: loss_func(x, **kwargs)                            
    
    if method is "momentum":                
    
        for i in iterator:
            
            idx = generate_index(response=response, batch_prop=batch_prop, 
                                 randomization=randomization, 
                                 seed=i)
            
            def objective(double[:] x):
                return loss_func(x, row_index=idx, **kwargs)
            
            #grad_i = numerical_gradient(objective, x)                        
            grad_i = calc_grad(objective, x)    
            
            diff = -np.asarray(x)

            if verbose == 2:
                print(f"\n x prev: {np.asarray(x)}")

            for j in range(p):                
                velocity[j] = mass * velocity[j] - learning_rate * grad_i[j]                        
                x[j] = x[j] + velocity[j]
            
            diff += np.asarray(x)
            
            if verbose == 2:
                print(f"\n x new: {np.asarray(x)}")

            losses.append(f(x))

            if (len(losses) > 3) and (np.abs(np.diff(losses[-2:])[0]) < tolerance):
                break
            
            if verbose == 2:       

                print("\n")
                print(f"iter {i+1} - decrease -----")

                try:
                    print(np.linalg.norm(diff, 1))                            
                except:
                    pass  

                print(f"iter {i+1} - loss -----")
                print(np.flip(losses)[0])

    
    if method in ("exp", "poly"):
        
        for i in iterator:
            
            idx = generate_index(response=response, batch_prop=batch_prop, 
                                 randomization=randomization, 
                                 seed=i)
            
            def objective(double[:] x):
                return loss_func(x, row_index=idx, **kwargs)
            
            #grad_i = numerical_gradient(objective, x)
            grad_i = calc_grad(objective, x)            
                        
            decay_rate = (1 + decay*i) if method is "poly" else exp(decay*i)
            
            diff = -np.asarray(x)

            if verbose == 2:
                print(f"\n x prev: {np.asarray(x)}")
            
            for j in range(p):   

                x[j] = x[j] - grad_i[j]*learning_rate/decay_rate

            if verbose == 2:
                print(f"\n x new: {np.asarray(x)}")

            diff += np.asarray(x)  
            
            losses.append(f(x))

            if (len(losses) > 3) and (np.abs(np.diff(losses[-2:])[0]) < tolerance):
                break
        
            if verbose == 2:

                print("\n")
                print(f"iter {i+1} - decrease -----")
                
                try:
                    print(np.linalg.norm(diff, 1))   
                except:
                    pass
                
                print(f"iter {i+1} - loss -----")
                print(np.flip(losses)[0])
    
    return np.asarray(x), num_iters, losses




# Authors: Thierry Moudiki
#
# License: BSD 3 Clear


import numpy as np
from sklearn.base import BaseEstimator
from . import _optimizerc as optimizerc


class Optimizer(BaseEstimator):
    """Optimizer class
    
    Attributes: 
    
        type_optim: str
            type of optimizer, (currently) either 'sgd' (stochastic minibatch gradient descent)
            or 'scd' (stochastic minibatch coordinate descent)

        num_iters: int
            number of iterations of the optimizer

        learning_rate: float
            step size

        batch_prop: float
            proportion of the initial data used at each optimization step

        learning_method: str
            "poly" - learning rate decreasing as a polynomial function 
            of # of iterations (default)
            "exp" - learning rate decreasing as an exponential function 
            of # of iterations
            "momentum" - gradient descent using momentum

        randomization: str
            type of randomization applied at each step
            "strat" - stratified subsampling (default)
            "shuffle" - random subsampling

        mass: float
            mass on velocity, for `method` == "momentum"

        decay: float
            coefficient of decrease of the learning rate for 
            `method` == "poly" and `method` == "exp"
            
        verbose: int 
            controls verbosity of gradient descent
            0 - nothing is printed
            1 - a progress bar is printed
            2 - succesive loss function values are printed

    """
    
    
    # construct the object -----

    def __init__(
        self,        
        type_optim="sgd",
        num_iters=100,         
        learning_rate=0.01, 
        batch_prop=1.0,         
        learning_method="momentum", 
        randomization="strat", 
        mass=0.9, decay=0.1, 
        verbose=1):
        
        self.type_optim = type_optim
        self.num_iters = num_iters
        self.learning_rate = learning_rate
        self.batch_prop = batch_prop
        self.learning_method = learning_method
        self.randomization = randomization
        self.mass = mass
        self.decay = decay
        self.verbose = verbose
        self.opt = None
                                                          
    
    def fit(self, loss_func, response, x0, **kwargs):
        """Fit GLM model to training data (X, y).
        
        Args:
                        
            loss_func: loss function
            
            response: array-like, shape = [n_samples]
            target variable (used for subsampling) 
            
            x0: array-like, shape = [n_features]
                initial value provided to the optimizer
        
            **kwargs: additional parameters to be passed to 
                    loss function
               
        Returns: 

            self: object
            
        """
        
        
        if self.type_optim == "scd":
                    
            self.results = optimizerc.scd(loss_func, 
                            response=response, x=x0, 
                            num_iters=self.num_iters,
                            batch_prop=self.batch_prop, 
                            learning_rate=self.learning_rate, 
                            learning_method=self.learning_method, 
                            mass=self.mass, decay=self.decay,                     
                            randomization=self.randomization,             
                            verbose=self.verbose, 
                            **kwargs)            
            
            
        if self.type_optim == "sgd":
                                        
            self.results = optimizerc.sgd(loss_func, 
                            response=response, x=x0, 
                            num_iters=self.num_iters,
                            batch_prop=self.batch_prop, 
                            learning_rate=self.learning_rate, 
                            learning_method=self.learning_method, 
                            mass=self.mass, decay=self.decay,                     
                            randomization=self.randomization,             
                            verbose=self.verbose, 
                            **kwargs)    

        return self

    def one_hot_encode(self, y, n_classes):
        return optimizerc.one_hot_encode(y, n_classes)


# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
import sklearn.metrics as skm2
from .custom import Custom
from ..utils import matrixops as mo
from sklearn.base import ClassifierMixin


class CustomClassifier(Custom, ClassifierMixin):
    """Custom Classification model 
    
    Attributes:

        obj: object
            any object containing a method fit (obj.fit()) and a method predict 
            (obj.predict())

        n_hidden_features: int
            number of nodes in the hidden layer

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not 
            (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out 
            of the training

        direct_link: boolean
            indicates if the original predictors are included (True) in model's 
            fitting or not (False)

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0: 
                no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian 
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant). 
            Currently available: standardization ('std') or MinMax scaling ('minmax')

        col_sample: float
            percentage of covariates randomly chosen for training   

        row_sample: float
            percentage of rows chosen for training, by stratified bootstrapping    

        seed: int 
            reproducibility seed for nodes_sim=='uniform'

        backend: str
            "cpu" or "gpu" or "tpu"                
    
    """

    # construct the object -----

    def __init__(
        self,
        obj,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
        backend="cpu"
    ):

        super().__init__(
            obj=obj,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend
        )

        self.type_fit = "classification"

    def fit(self, X, y, sample_weight=None, **kwargs):
        """Fit custom model to training data (X, y).
        
        Args: 
    
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
            
            y: array-like, shape = [n_samples]
                Target values.

            **kwargs: additional parameters to be passed to 
                        self.cook_training_set or self.obj.fit
            
        Returns:
    
            self: object
        """

        output_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)

        # if sample_weights, else: (must use self.row_index)
        if sample_weight is not None:

            self.obj.fit(
                scaled_Z,
                output_y,
                sample_weight=np.ravel(sample_weight, order='C')[self.index_row],
                **kwargs
            )

            return self

        # if sample_weight is None:
        self.obj.fit(scaled_Z, output_y, **kwargs)

        return self

    def predict(self, X, **kwargs):
        """Predict test data X.
        
        Args:
    
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
    
            **kwargs: additional parameters to be passed to 
                    self.cook_test_set
            
        Returns:         

            model predictions: {array-like}
        """

        if len(X.shape) == 1:

            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            return (
                self.obj.predict(self.cook_test_set(new_X, **kwargs), **kwargs)
            )[0]

        return self.obj.predict(self.cook_test_set(X, **kwargs), **kwargs)

    def predict_proba(self, X, **kwargs):
        """Predict probabilities for test data X.
        
        Args: 
    
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
    
            **kwargs: additional parameters to be passed to 
                    self.cook_test_set
            
        Returns: 
    
            probability estimates for test data: {array-like}        
        """

        if len(X.shape) == 1:

            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            return (
                self.obj.predict_proba(
                    self.cook_test_set(new_X, **kwargs), **kwargs
                )
            )[0]

        return self.obj.predict_proba(self.cook_test_set(X, **kwargs), **kwargs)

    def score(self, X, y, scoring=None, **kwargs):
        """ Score the model on test set features X and response y. 

        Args:
    
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features

            y: array-like, shape = [n_samples]
                Target values

            scoring: str
                must be in ('explained_variance', 'neg_mean_absolute_error', \
                    'neg_mean_squared_error', 'neg_mean_squared_log_error', \
                    'neg_median_absolute_error', 'r2')
    
            **kwargs: 
                additional parameters to be passed to scoring functions
            
        Returns: 
    
            model scores: {array-like}

        """

        preds = self.predict(X)

        if scoring is None:
            scoring = "accuracy"

        # check inputs
        assert scoring in (
            "accuracy",
            "average_precision",
            "brier_score_loss",
            "f1",
            "f1_micro",
            "f1_macro",
            "f1_weighted",
            "f1_samples",
            "neg_log_loss",
            "precision",
            "recall",
            "roc_auc",
        ), "'scoring' should be in ('accuracy', 'average_precision', \
                           'brier_score_loss', 'f1', 'f1_micro', \
                           'f1_macro', 'f1_weighted',  'f1_samples', \
                           'neg_log_loss', 'precision', 'recall', \
                           'roc_auc')"

        scoring_options = {
            "accuracy": skm2.accuracy_score,
            "average_precision": skm2.average_precision_score,
            "brier_score_loss": skm2.brier_score_loss,
            "f1": skm2.f1_score,
            "f1_micro": skm2.f1_score,
            "f1_macro": skm2.f1_score,
            "f1_weighted": skm2.f1_score,
            "f1_samples": skm2.f1_score,
            "neg_log_loss": skm2.log_loss,
            "precision": skm2.precision_score,
            "recall": skm2.recall_score,
            "roc_auc": skm2.roc_auc_score,
        }

        return scoring_options[scoring](y, preds, **kwargs)



# adapted from sklearn

import sys
import os

from nnetsauce._build_utils import cythonize_extensions


def configuration(parent_package="", top_path=None):
    from numpy.distutils.misc_util import Configuration
    import numpy

    libraries = []
    if os.name == "posix":
        libraries.append("m")

    config = Configuration("nnetsauce", parent_package, top_path)

    # submodules with build utilities
    config.add_subpackage("__check_build")
    config.add_subpackage("_build_utils")

    # submodules which do not have their own setup.py
    # we must manually add sub-submodules & tests
    config.add_subpackage("demo")
    config.add_subpackage("utils")
    config.add_subpackage("base")
    config.add_subpackage("boosting")
    config.add_subpackage("custom")
    config.add_subpackage("glm")    
    config.add_subpackage("mts")
    config.add_subpackage("multitask")
    config.add_subpackage("optimizers")
    config.add_subpackage("randombag")
    config.add_subpackage("ridge2")
    config.add_subpackage("rvfl")        
    config.add_subpackage("simulation")

    # submodules which have their own setup.py
    config.add_subpackage("optimizers")
    config.add_subpackage("sampling")    

    # add the test directory
    config.add_subpackage("tests")

    # Skip cythonization as we do not want to include the generated
    # C/C++ files in the release tarballs as they are not necessarily
    # forward compatible with future versions of Python for instance.
    if "sdist" not in sys.argv:
        cythonize_extensions(top_path, config)

    return config


if __name__ == "__main__":
    from numpy.distutils.core import setup

    setup(**configuration(top_path="").todict())



"""Check if vector is factor."""

# Authors: T. Moudiki <thierry.moudiki@gmail.com>
#
# License: BSD 3 Clause Clear

import ctypes
import numpy as np
import os
from ctypes import c_double


dir_path = os.path.dirname(__file__)
try:
    check_factorer = ctypes.cdll.LoadLibrary(dir_path + "/check_factor.so")
except:
    check_factorer = ctypes.CDLL(dir_path + "/check_factor.so")


def is_factor2(x):

    x_ = x.tolist() if isinstance(x, np.ndarray) else x.copy()

    n = len(x_)
    x_c = (c_double * n)(*x_)  # Create ctypes pointer to underlying memory

    return check_factorer.check_factor(x_c, n)


# if __name__=="__main__":
#   x1 = [65, 66, 67, 68.5, 69]
#   x2 = [65, 66, 67, 68]
#   print(check_factor(x1))
#   print(check_factor(x2))



from sys import version_info as _swig_python_version_info

if _swig_python_version_info < (2, 7, 0):
    raise RuntimeError("Python 2.7 or later required")

    # Import the low-level C/C++ module
    # if __package__ or "." in __name__:

    try:

        from ._sobol import *

    except:

        import _sobol

    else:

        try:
            import builtins as __builtin__
        except ImportError:
            import __builtin__

        def _swig_repr(self):
            try:
                strthis = "proxy of " + self.this.__repr__()
            except __builtin__.Exception:
                strthis = ""
            return "<%s.%s; %s >" % (
                self.__class__.__module__,
                self.__class__.__name__,
                strthis,
            )

        def _swig_setattr_nondynamic_instance_variable(set):
            def set_instance_attr(self, name, value):
                if name == "thisown":
                    self.this.own(value)
                elif name == "this":
                    set(self, name, value)
                elif hasattr(self, name) and isinstance(
                    getattr(type(self), name), property
                ):
                    set(self, name, value)
                else:
                    raise AttributeError(
                        "You cannot add instance attributes to %s" % self
                    )

            return set_instance_attr

        def _swig_setattr_nondynamic_class_variable(set):
            def set_class_attr(cls, name, value):
                if hasattr(cls, name) and not isinstance(
                    getattr(cls, name), property
                ):
                    set(cls, name, value)
                else:
                    raise AttributeError(
                        "You cannot add class attributes to %s" % cls
                    )

            return set_class_attr

        def _swig_add_metaclass(metaclass):
            """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""

            def wrapper(cls):
                return metaclass(
                    cls.__name__, cls.__bases__, cls.__dict__.copy()
                )

            return wrapper

        class _SwigNonDynamicMeta(type):
            """Meta class to enforce nondynamic attributes (no new attributes) for a class"""

            __setattr__ = _swig_setattr_nondynamic_class_variable(
                type.__setattr__
            )

        def i8_bit_hi1(n):
            return _sobol.i8_bit_hi1(n)

        def i8_bit_lo0(n):
            return _sobol.i8_bit_lo0(n)

        def i8_max(i1, i2):
            return _sobol.i8_max(i1, i2)

        def i8_min(i1, i2):
            return _sobol.i8_min(i1, i2)

        def i8_sobol(dim_num, seed, quasi):
            return _sobol.i8_sobol(dim_num, seed, quasi)

        def i8_sobol_generate(m, n, skip):
            return _sobol.i8_sobol_generate(m, n, skip)

        def i8_uniform(b, c, seed):
            return _sobol.i8_uniform(b, c, seed)

        def r8_abs(x):
            return _sobol.r8_abs(x)

        def r8_nint(x):
            return _sobol.r8_nint(x)

        def r8_uniform_01(seed):
            return _sobol.r8_uniform_01(seed)

        def tau_sobol(dim_num):
            return _sobol.tau_sobol(dim_num)

        def timestamp():
            return _sobol.timestamp()



"""
-- adapted from sklearn
"""
import sys
import logging
import os

from ._config import get_config, set_config, config_context

logger = logging.getLogger(__name__)


# PEP0440 compatible formatted version, see:
# https://www.python.org/dev/peps/pep-0440/
#
# Generic release markers:
#   X.Y
#   X.Y.Z   # For bugfix releases
#
# Admissible pre-release markers:
#   X.YaN   # Alpha release
#   X.YbN   # Beta release
#   X.YrcN  # Release Candidate
#   X.Y     # Final release
#
# Dev branch marker is: 'X.Y.dev' or 'X.Y.devN' where N is an integer.
# 'X.Y.dev0' is the canonical version of 'X.Y.dev'
#
__version__ = "0.5.2"


# On OSX, we can get a runtime error due to multiple OpenMP libraries loaded
# simultaneously. This can happen for instance when calling BLAS inside a
# prange. Setting the following environment variable allows multiple OpenMP
# libraries to be loaded. It should not degrade performances since we manually
# take care of potential over-subcription performance issues, in sections of
# the code where nested OpenMP loops can happen, by dynamically reconfiguring
# the inner OpenMP runtime to temporarily disable it while under the scope of
# the outer OpenMP parallel section.
os.environ.setdefault("KMP_DUPLICATE_LIB_OK", "True")

# Workaround issue discovered in intel-openmp 2019.5:
# https://github.com/ContinuumIO/anaconda-issues/issues/11294
os.environ.setdefault("KMP_INIT_AT_FORK", "FALSE")

try:
    # This variable is injected in the __builtins__ by the build
    # process. It is used to enable importing subpackages of nnetsauce when
    # the binaries are not built
    # mypy error: Cannot determine type of '__NNETSAUCE_SETUP__'
    __NNETSAUCE_SETUP__  # type: ignore
except NameError:
    __NNETSAUCE_SETUP__ = False

if __NNETSAUCE_SETUP__:

    sys.stderr.write("Partial import of nnetsauce during the build process.\n")
    # We are not importing the rest of scikit-learn during the build
    # process, as it may not be compiled yet

else:

    from .base.base import Base
    from .base.baseRegressor import BaseRegressor
    from .boosting.adaBoostClassifier import AdaBoostClassifier
    from .custom.customClassifier import CustomClassifier
    from .custom.customRegressor import CustomRegressor
    from .glm.glmClassifier import GLMClassifier
    from .glm.glmRegressor import GLMRegressor
    from .mts.mts import MTS
    from .multitask.multitaskClassifier import MultitaskClassifier
    from .optimizers._optimizer import Optimizer
    from .randombag.randomBagClassifier import RandomBagClassifier
    from .ridge2.ridge2Classifier import Ridge2Classifier
    from .ridge2.ridge2Regressor import Ridge2Regressor
    from .ridge2.ridge2MultitaskClassifier import Ridge2MultitaskClassifier
    # from .rnn.rnnRegressor import RNNRegressor
    # from .rnn.rnnClassifier import RNNClassifier
    from .rvfl.bayesianrvflRegressor import BayesianRVFLRegressor
    from .rvfl.bayesianrvfl2Regressor import BayesianRVFL2Regressor

    __all__ = [
        "AdaBoostClassifier",
        "Base",
        "BaseRegressor",
        "BayesianRVFLRegressor",
        "BayesianRVFL2Regressor",
        "CustomClassifier",
        "CustomRegressor",
        "GLMRegressor", 
        "GLMClassifier",
        "MultitaskClassifier",
        "Optimizer",
        "RandomBagClassifier",
        "Ridge2Regressor",
        "Ridge2Classifier",
        "Ridge2MultitaskClassifier",
        #    "RNNRegressor",
        #    "RNNClassifier",
        "MTS",
    ]

def setup_module(module):
    """Fixture for the tests to assure globally controllable seeding of RNGs"""
    import os
    import numpy as np
    import random

    # Check if a random seed exists in the environment, if not create one.
    _random_seed = os.environ.get("NNETSAUCE_SEED", None)
    if _random_seed is None:
        _random_seed = np.random.uniform() * np.iinfo(np.int32).max
    _random_seed = int(_random_seed)
    print("I: Seeding RNGs with %r" % _random_seed)
    np.random.seed(_random_seed)
    random.seed(_random_seed)




from . import _rowsubsamplingc as rowsubsamplingc


class SubSampler():
    """Subsampling class.
        
     Attributes: 
     
        y: array-like, shape = [n_samples]
            Target values.

        row_sample: double
            subsampling fraction

        seed: int
            reproductibility seed   

    """

    def __init__(self, y, row_sample=0.8, seed=123):
        self.y = y
        self.row_sample = row_sample
        self.seed = seed

    def subsample(self):
        return rowsubsamplingc.subsamplec(self.y, self.row_sample, self.seed)


import numpy as np
import platform
from .memoize import memoize
from jax import device_put
from scipy import sparse
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
if platform.system() in ('Linux', 'Darwin'):
    import jax.numpy as jnp


# column bind
def cbind(x, y, backend="cpu"):
    # if len(x.shape) == 1 or len(y.shape) == 1:
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ('Linux', 'Darwin')):
        return jnp.column_stack((x, y))
    return np.column_stack((x, y))


# center... response
@memoize
def center_response(y):
    y_mean = np.mean(y)
    return y_mean, (y - y_mean)


# cluster the covariates
def cluster_covariates(X, n_clusters, seed, type_clust="kmeans", **kwargs):

    if type_clust == "kmeans":
        kmeans = KMeans(n_clusters=n_clusters, random_state=seed, **kwargs)
        kmeans.fit(X)

        return kmeans, kmeans.predict(X)

    elif type_clust == "gmm":
        gmm = GaussianMixture(
            n_components=n_clusters, random_state=seed, **kwargs
        )
        gmm.fit(X)

        return gmm, gmm.predict(X)


# computes t(x)%*%y
def crossprod(x, y=None, backend="cpu"):
    # assert on dimensions
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ('Linux', 'Darwin')):
        x = device_put(x)
        if y is None:
            return jnp.dot(x.T, x).block_until_ready()
        y = device_put(y)
        return jnp.dot(x.T, y).block_until_ready()
    if y is None:
        return np.dot(x.transpose(), x)
    return np.dot(x.transpose(), y)


# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# dropout
def dropout(x, drop_prob=0, seed=123):

    assert 0 <= drop_prob <= 1

    n, p = x.shape

    if drop_prob == 0:
        return x

    if drop_prob == 1:
        return np.zeros_like(x)

    np.random.seed(seed)
    dropped_indices = np.random.rand(n, p) > drop_prob

    return dropped_indices * x / (1 - drop_prob)


# one-hot encoding
def one_hot_encode(x_clusters, n_clusters):

    assert (
        max(x_clusters) <= n_clusters
    ), "you must have max(x_clusters) <= n_clusters"

    n_obs = len(x_clusters)
    res = np.zeros((n_obs, n_clusters))

    for i in range(n_obs):
        res[i, x_clusters[i]] = 1

    return res


# one-hot encoding
def one_hot_encode2(y, n_classes):

    n_obs = len(y)
    res = np.zeros((n_obs, n_classes))

    for i in range(n_obs):
        res[i, y[i]] = 1

    return res


# row bind
def rbind(x, y, backend="cpu"):
    # if len(x.shape) == 1 or len(y.shape) == 1:
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ('Linux', 'Darwin')):
        return jnp.row_stack((x, y))
    return np.row_stack((x, y))


# adapted from sklearn.utils.exmath
def safe_sparse_dot(a, b, backend="cpu", dense_output=False):
    """Dot product that handle the sparse matrix case correctly

    Parameters
    ----------
    a : array or sparse matrix
    b : array or sparse matrix
    dense_output : boolean, (default=False)
        When False, ``a`` and ``b`` both being sparse will yield sparse output.
        When True, output will always be a dense array.

    Returns
    -------
    dot_product : array or sparse matrix
        sparse if ``a`` and ``b`` are sparse and ``dense_output=False``.
    """
    sys_platform = platform.system()

    if backend in ("gpu", "tpu") and (sys_platform in ('Linux', 'Darwin')):
        # modif when jax.scipy.sparse available
        return jnp.dot(device_put(a), device_put(b)).block_until_ready()

    #    if backend == "cpu":
    if a.ndim > 2 or b.ndim > 2:
        if sparse.issparse(a):
            # sparse is always 2D. Implies b is 3D+
            # [i, j] @ [k, ..., l, m, n] -> [i, k, ..., l, n]
            b_ = np.rollaxis(b, -2)
            b_2d = b_.reshape((b.shape[-2], -1))
            ret = a @ b_2d
            ret = ret.reshape(a.shape[0], *b_.shape[1:])
        elif sparse.issparse(b):
            # sparse is always 2D. Implies a is 3D+
            # [k, ..., l, m] @ [i, j] -> [k, ..., l, j]
            a_2d = a.reshape(-1, a.shape[-1])
            ret = a_2d @ b
            ret = ret.reshape(*a.shape[:-1], b.shape[1])
        else:
            ret = np.dot(a, b)
    else:
        ret = a @ b

    if (
        sparse.issparse(a)
        and sparse.issparse(b)
        and dense_output
        and hasattr(ret, "toarray")
    ):
        return ret.toarray()

    return ret

# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# Obtain this for JAX
# scale... covariates
def scale_covariates(X, choice="std", training=True, scaler=None):

    scaling_options = {
        "std": StandardScaler(copy=True, with_mean=True, with_std=True),
        "minmax": MinMaxScaler(),
    }

    if training == True:
        # scaler must be not None
        scaler = scaling_options[choice]
        scaled_X = scaler.fit_transform(X)
        return scaler, scaled_X

    # training == False:
    # scaler must be not None
    return scaler.transform(X)


# from sklearn.utils.exmath
def squared_norm(x, backend="cpu"):
    """Squared Euclidean or Frobenius norm of x.

    Faster than norm(x) ** 2.

    Parameters
    ----------
    x : array_like

    Returns
    -------
    float
        The Euclidean norm when x is a vector, the Frobenius norm when x
        is a matrix (2-d array).
    """
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ('Linux', 'Darwin')):
        x = np.ravel(x, order="K")
        x = device_put(x)        
        return jnp.dot(x, x).block_until_ready()

    x = np.ravel(x, order="K")
    return np.dot(x, x)


# computes x%*%t(y)
def tcrossprod(x, y=None, backend="cpu"):
    # assert on dimensions
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ('Linux', 'Darwin')):
        x = device_put(x)
        if y is None:
            return jnp.dot(x, x.T).block_until_ready()
        y = device_put(y)
        return jnp.dot(x, y.T).block_until_ready()
    if y is None:
        return np.dot(x, x.transpose())
    return np.dot(x, y.transpose())


# convert vector to numpy array
def to_np_array(X):
    return np.array(X.copy(), ndmin=2)



import numpy as np


def relu(x):
    return np.maximum(x, 0)


def sigmoid(x):
    return np.exp(-np.logaddexp(0, -x))


def prelu(x, a=None):

    if a is not None:
        y = x.copy()
        index = x < 0
        y[index] = a * x[index]
        return y
    else:
        raise NotImplementedError


def elu(x, a=None):

    if a is not None:
        y = x.copy()
        index = x < 0
        y[index] = a * (np.exp(x[index]) - 1)
        return y
    else:
        raise NotImplementedError



# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
import pickle
import sklearn.metrics as skm2
from ..utils import matrixops as mo
from ..utils import misc as mx
from ..base import Base
from scipy.special import expit
from sklearn.base import ClassifierMixin


class MultitaskClassifier(Base, ClassifierMixin):
    """Multitask Classification model based on regression models, with shared covariates
    
    Attributes: 
    
        obj: object
            any object (must be a regression model) containing a method fit (obj.fit()) 
            and a method predict (obj.predict())

        n_hidden_features: int
            number of nodes in the hidden layer

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not 
            (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out 
            of the training

        direct_link: boolean
            indicates if the original predictors are included (True) in model's 
            fitting or not (False)

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0: 
                no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian 
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant). 
            Currently available: standardization ('std') or MinMax scaling ('minmax')

        col_sample: float
            percentage of covariates randomly chosen for training   

        row_sample: float
            percentage of rows chosen for training, by stratified bootstrapping    

        seed: int 
            reproducibility seed for nodes_sim=='uniform'

        backend: str
            "cpu" or "gpu" or "tpu"                           

    References:       

        [1] Moudiki, T. (2020). Quasi-randomized networks for regression and classification, with two shrinkage parameters. Available at: 
        https://www.researchgate.net/publication/339512391_Quasi-randomized_networks_for_regression_and_classification_with_two_shrinkage_parameters    
       
    """

    # construct the object -----

    def __init__(
        self,
        obj,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
        backend="cpu"
    ):

        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend
        )

        self.type_fit = "classification"
        self.obj = obj
        self.fit_objs = {}

    def fit(self, X, y, sample_weight=None, **kwargs):
        """Fit MultitaskClassifier to training data (X, y).
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
            
            y: array-like, shape = [n_samples]
                Target values.
        
            **kwargs: additional parameters to be passed to 
                    self.cook_training_set or self.obj.fit
               
        Returns:
        
            self: object

        """

        assert mx.is_factor(y), "y must contain only integers"

        output_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)

        self.n_classes = len(np.unique(y))

        # multitask response
        Y = mo.one_hot_encode2(output_y, self.n_classes)

        # if sample_weight is None:
        for i in range(self.n_classes):

            self.fit_objs[i] = pickle.loads(
                pickle.dumps(self.obj.fit(scaled_Z, Y[:, i], **kwargs), -1)
            )

        return self

    def predict(self, X, **kwargs):
        """Predict test data X.
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
            
            **kwargs: additional parameters to be passed to 
                    self.cook_test_set
               
        Returns:
        
            model predictions: {array-like}

        """

        return np.argmax(self.predict_proba(X, **kwargs), axis=1)

    def predict_proba(self, X, **kwargs):
        """Predict probabilities for test data X.
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
            
            **kwargs: additional parameters to be passed to 
                    self.cook_test_set
               
        Returns:
        
            probability estimates for test data: {array-like}   
                 
        """

        shape_X = X.shape

        probs = np.zeros((shape_X[0], self.n_classes))

        if len(shape_X) == 1:

            n_features = shape_X[0]

            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            Z = self.cook_test_set(new_X, **kwargs)

            # loop on all the classes
            for i in range(self.n_classes):

                probs[:, i] = self.fit_objs[i].predict(Z, **kwargs)[0]

        else:

            Z = self.cook_test_set(X, **kwargs)

            # loop on all the classes
            for i in range(self.n_classes):

                probs[:, i] = self.fit_objs[i].predict(Z, **kwargs)

        expit_raw_probs = expit(probs)

        return expit_raw_probs / expit_raw_probs.sum(axis=1)[:, None]

    def score(self, X, y, scoring=None, **kwargs):
        """ Score the model on test set features X and response y. 

        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features

            y: array-like, shape = [n_samples]
                Target values

            scoring: str
                must be in ('explained_variance', 'neg_mean_absolute_error', \
                            'neg_mean_squared_error', 'neg_mean_squared_log_error', \
                            'neg_median_absolute_error', 'r2')
            
            **kwargs: additional parameters to be passed to scoring functions
               
        Returns: 
        
            model scores: {array-like}

        """

        preds = self.predict(X)

        if scoring is None:
            scoring = "accuracy"

        # check inputs
        assert scoring in (
            "accuracy",
            "average_precision",
            "brier_score_loss",
            "f1",
            "f1_micro",
            "f1_macro",
            "f1_weighted",
            "f1_samples",
            "neg_log_loss",
            "precision",
            "recall",
            "roc_auc",
        ), "'scoring' should be in ('accuracy', 'average_precision', \
                           'brier_score_loss', 'f1', 'f1_micro', \
                           'f1_macro', 'f1_weighted',  'f1_samples', \
                           'neg_log_loss', 'precision', 'recall', \
                           'roc_auc')"

        scoring_options = {
            "accuracy": skm2.accuracy_score,
            "average_precision": skm2.average_precision_score,
            "brier_score_loss": skm2.brier_score_loss,
            "f1": skm2.f1_score,
            "f1_micro": skm2.f1_score,
            "f1_macro": skm2.f1_score,
            "f1_weighted": skm2.f1_score,
            "f1_samples": skm2.f1_score,
            "neg_log_loss": skm2.log_loss,
            "precision": skm2.precision_score,
            "recall": skm2.recall_score,
            "roc_auc": skm2.roc_auc_score,
        }

        return scoring_options[scoring](y, preds, **kwargs)



import numpy as np
from .memoize import memoize


# merge two dictionaries
def merge_two_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z


# check if x is int
@memoize
def is_int(x):
    try:
        return int(x) == x
    except:
        return False


# check if x is float
@memoize
def is_float(x):
    return isinstance(x, float)


# check if the response contains only integers
@memoize
def is_factor(y):
    return not np.mod(y, 1).any()


# flatten list of lists
flatten = lambda l: [item for sublist in l for item in sublist]



# Authors: Thierry Moudiki
#
# License: BSD 3 Clear

import numpy as np
import platform
import sys
import sklearn.metrics as skm2
from scipy.optimize import minimize
from .ridge2 import Ridge2
from ..utils import matrixops as mo
from ..utils import misc as mx
from sklearn.base import ClassifierMixin
from scipy.special import logsumexp


class Ridge2Classifier(Ridge2, ClassifierMixin):
    """Multinomial logit classification with 2 regularization parameters
    
    Attributes: 
    
        n_hidden_features: int
            number of nodes in the hidden layer

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not 
            (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out 
            of the training

        direct_link: boolean
            indicates if the original predictors are included (True) in model's 
            fitting or not (False)

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0: 
                no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian 
            Mixture Model ('gmm')
            
        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant). 
            Currently available: standardization ('std') or MinMax scaling ('minmax')

        col_sample: float
            percentage of covariates randomly chosen for training   

        row_sample: float
            percentage of rows chosen for training, by stratified bootstrapping    

        lambda1: float
            regularization parameter on direct link

        lambda2: float
            regularization parameter on hidden layer

        seed: int 
            reproducibility seed for nodes_sim=='uniform'

        backend: str
            "cpu" or "gpu" or "tpu"                


    References:
    
        - [1] Moudiki, T. (2020). Quasi-randomized networks for regression and classification, with two shrinkage parameters. Available at: 
        https://www.researchgate.net/publication/339512391_Quasi-randomized_networks_for_regression_and_classification_with_two_shrinkage_parameters

        - [2] Moudiki, T. (2019). Multinomial logistic regression using quasi-randomized networks. Available at: 
        https://www.researchgate.net/publication/334706878_Multinomial_logistic_regression_using_quasi-randomized_networks 

    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        lambda1=0.1,
        lambda2=0.1,
        seed=123,
        backend="cpu"
    ):

        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            lambda1=lambda1,
            lambda2=lambda2,
            seed=seed,
            backend=backend 
        )


        self.type_fit = "classification"

    def loglik(self, X, Y, **kwargs):
        """Log-likelihood for training data (X, Y).
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
            
            Y: array-like, shape = [n_samples]
                One-hot encode target values.
        
            **kwargs: additional parameters to be passed to 
                    self.cook_training_set or self.obj.fit
               
        Returns: 
        
        """

        def loglik_grad_hess(Y, X, B, XB, hessian=True, **kwargs):
            # nobs, n_classes
            n, K = Y.shape

            # total number of covariates
            p = X.shape[1]

            # initial number of covariates
            init_p = p - self.n_hidden_features

            max_double = 709.0
            XB[XB > max_double] = max_double
            exp_XB = np.exp(XB)
            probs = exp_XB / exp_XB.sum(axis=1)[:, None]

            # gradient -----
            # (Y - p) -> (n, K)
            # X -> (n, p)
            # (K, n) %*% (n, p) -> (K, p)
            if hessian is False:
                grad = (
                    -mo.safe_sparse_dot(
                        a=(Y - probs).T, b=X, backend=self.backend
                    )
                    / n
                )
                grad += self.lambda1 * B[0:init_p, :].sum(axis=0)[:, None]
                grad += self.lambda2 * B[init_p:p, :].sum(axis=0)[:, None]

                return grad.flatten()

            # hessian -----
            if hessian is True:
                Kp = K * p
                hess = np.zeros((Kp, Kp), float)
                for k1 in range(K):
                    x_index = range(k1 * p, (k1 + 1) * p)
                    for k2 in range(k1, K):
                        y_index = range(k2 * p, (k2 + 1) * p)
                        H_sub = (
                            -mo.safe_sparse_dot(
                                a=X.T,
                                b=(probs[:, k1] * probs[:, k2])[:, None] * X,
                                backend=self.backend,
                            )
                            / n
                        )  # do not store
                        hess[np.ix_(x_index, y_index)] = hess[
                            np.ix_(y_index, x_index)
                        ] = H_sub

                return hess + (self.lambda1 + self.lambda2) * np.identity(Kp)

        # total number of covariates
        p = X.shape[1]

        # initial number of covariates
        init_p = p - self.n_hidden_features

        # log-likelihood (1st return)
        def loglik_func(x):
            # (p, K)
            B = x.reshape(Y.shape[1], p).T

            # (n, K)
            XB = mo.safe_sparse_dot(X, B, backend=self.backend)

            res = -(np.sum(Y * XB, axis=1) - logsumexp(XB)).mean()

            res += 0.5 * self.lambda1 * mo.squared_norm(B[0:init_p, :], backend=self.backend)
            res += 0.5 * self.lambda2 * mo.squared_norm(B[init_p:p, :], backend=self.backend)

            return res

        # gradient of log-likelihood
        def grad_func(x):
            # (p, K)
            B = x.reshape(Y.shape[1], p).T

            return loglik_grad_hess(
                Y=Y,
                X=X,
                B=B,
                XB=mo.safe_sparse_dot(X, B, backend=self.backend),
                hessian=False,
                **kwargs
            )

        # hessian of log-likelihood
        def hessian_func(x):
            # (p, K)
            B = x.reshape(Y.shape[1], p).T

            return loglik_grad_hess(
                Y=Y,
                X=X,
                B=B,
                XB=mo.safe_sparse_dot(X, B, backend=self.backend),
                hessian=True,
                **kwargs
            )

        return loglik_func, grad_func, hessian_func

    # newton-cg
    # L-BFGS-B
    def fit(self, X, y, solver="L-BFGS-B", **kwargs):
        """Fit Ridge model to training data (X, y).
           
        for beta: regression coeffs (beta11, ..., beta1p, ..., betaK1, ..., betaKp)
        for K classes and p covariates.
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
            
            y: array-like, shape = [n_samples]
                Target values.
        
            **kwargs: additional parameters to be passed to 
                    self.cook_training_set or self.obj.fit
               
        Returns: 
        
            self: object

        """

        assert mx.is_factor(y), "y must contain only integers"

        output_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)

        self.n_classes = len(np.unique(y))

        Y = mo.one_hot_encode2(output_y, self.n_classes)

        # optimize for beta, minimize self.loglik (maximize loglik) -----
        loglik_func, grad_func, hessian_func = self.loglik(X=scaled_Z, Y=Y)

        if solver == "L-BFGS-B":
            self.beta = minimize(
                fun=loglik_func,
                x0=np.zeros(scaled_Z.shape[1] * self.n_classes),
                jac=grad_func,
                method=solver,
            ).x

        if solver in ("Newton-CG", "trust-ncg"):
            self.beta = minimize(
                fun=loglik_func,
                x0=np.zeros(scaled_Z.shape[1] * self.n_classes),
                jac=grad_func,
                hess=hessian_func,
                method=solver,
            ).x

        return self

    def predict(self, X, **kwargs):
        """Predict test data X.
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
            
            **kwargs: additional parameters to be passed to 
                    self.cook_test_set
               
        Returns: 
        
            model predictions: {array-like}
        """

        return np.argmax(self.predict_proba(X, **kwargs), axis=1)

    def predict_proba(self, X, **kwargs):
        """Predict probabilities for test data X.
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
            
            **kwargs: additional parameters to be passed to 
                    self.cook_test_set
               
        Returns: 
        
            probability estimates for test data: {array-like}    
                
        """
        if len(X.shape) == 1:

            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            Z = self.cook_test_set(new_X, **kwargs)

        else:

            Z = self.cook_test_set(X, **kwargs)

        ZB = mo.safe_sparse_dot(
            a=Z,
            b=self.beta.reshape(
                self.n_classes,
                X.shape[1] + self.n_hidden_features + self.n_clusters).T,        
            backend=self.backend)
        
        exp_ZB = np.exp(ZB)

        return exp_ZB / exp_ZB.sum(axis=1)[:, None]

    def score(self, X, y, scoring=None, **kwargs):
        """ Score the model on test set features X and response y. 

        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features

            y: array-like, shape = [n_samples]
                Target values

            scoring: str
                must be in ('explained_variance', 'neg_mean_absolute_error', \
                            'neg_mean_squared_error', 'neg_mean_squared_log_error', \
                            'neg_median_absolute_error', 'r2')
            
            **kwargs: additional parameters to be passed to scoring functions
               
        Returns: 
        
            model scores: {array-like}

        """

        preds = self.predict(X)

        if scoring is None:
            scoring = "accuracy"

        # check inputs
        assert scoring in (
            "accuracy",
            "average_precision",
            "brier_score_loss",
            "f1",
            "f1_micro",
            "f1_macro",
            "f1_weighted",
            "f1_samples",
            "neg_log_loss",
            "precision",
            "recall",
            "roc_auc",
        ), "'scoring' should be in ('accuracy', 'average_precision', \
                           'brier_score_loss', 'f1', 'f1_micro', \
                           'f1_macro', 'f1_weighted',  'f1_samples', \
                           'neg_log_loss', 'precision', 'recall', \
                           'roc_auc')"

        scoring_options = {
            "accuracy": skm2.accuracy_score,
            "average_precision": skm2.average_precision_score,
            "brier_score_loss": skm2.brier_score_loss,
            "f1": skm2.f1_score,
            "f1_micro": skm2.f1_score,
            "f1_macro": skm2.f1_score,
            "f1_weighted": skm2.f1_score,
            "f1_samples": skm2.f1_score,
            "neg_log_loss": skm2.log_loss,
            "precision": skm2.precision_score,
            "recall": skm2.recall_score,
            "roc_auc": skm2.roc_auc_score,
        }

        return scoring_options[scoring](y, preds, **kwargs)



"""Find index of element in vector."""

# Authors: T. Moudiki <thierry.moudiki@gmail.com>
#
# License: BSD 3 Clause Clear

import ctypes
import numpy as np
import os
from ctypes import c_double, c_long


dir_path = os.path.dirname(__file__)
try:
    wherer = ctypes.cdll.LoadLibrary(dir_path + "/wherer.so")
except:
    wherer = ctypes.CDLL(dir_path + "/wherer.so")


def index_where(x, elt):

    x_ = x.tolist() if isinstance(x, np.ndarray) else x.copy()

    n = len(x_)
    x_c = (c_double * n)(*x_)  # Create ctypes pointer to underlying memory
    res = (c_long * n)(*([0] * n))

    wherer.where(x_c, c_double(elt), n, res)

    z = np.asarray(list(res))

    return z[z != -1]


# if __name__=="__main__":
#   x = [65, 68, 68, 66, 67, 68, 75, 76, 68, 68.1, 68]
#   print(index_where(x, 68))




# Authors: Thierry Moudiki
#
# License: BSD 3 Clear


import pickle 
import numpy as np
import sklearn.metrics as skm2
from .glm import GLM
from ..utils import matrixops as mo
from ..utils import misc as mx
from sklearn.base import RegressorMixin
from scipy.optimize import minimize
from ..optimizers import Optimizer
from scipy.special import erf, factorial


class GLMRegressor(GLM, RegressorMixin):
    """Generalized 'linear' models using quasi-randomized networks (regression)
        
        Attributes:
        
            n_hidden_features: int
                number of nodes in the hidden layer

            lambda1: float
                regularization parameter for GLM coefficients on original features

            alpha1: float
                controls compromize between l1 and l2 norm of GLM coefficients on original features

            lambda2: float
                regularization parameter for GLM coefficients on nonlinear features

            alpha2: float
                controls compromize between l1 and l2 norm of GLM coefficients on nonlinear features

            activation_name: str
                activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

            a: float
                hyperparameter for 'prelu' or 'elu' activation function

            nodes_sim: str
                type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
                'uniform'

            bias: boolean
                indicates if the hidden layer contains a bias term (True) or not 
                (False)

            dropout: float
                regularization parameter; (random) percentage of nodes dropped out 
                of the training

            direct_link: boolean
                indicates if the original predictors are included (True) in model's 
                fitting or not (False)

            n_clusters: int
                number of clusters for 'kmeans' or 'gmm' clustering (could be 0: 
                    no clustering)

            cluster_encode: bool
                defines how the variable containing clusters is treated (default is one-hot)
                if `False`, then labels are used, without one-hot encoding

            type_clust: str
                type of clustering method: currently k-means ('kmeans') or Gaussian 
                Mixture Model ('gmm')
                
            type_scaling: a tuple of 3 strings
                scaling methods for inputs, hidden layer, and clustering respectively
                (and when relevant). 
                Currently available: standardization ('std') or MinMax scaling ('minmax')

            optimizer: object 
                optimizer, from class nnetsauce.utils.Optimizer

            seed: int 
                reproducibility seed for nodes_sim=='uniform'

    """        
    
    # construct the object -----

    def __init__(
        self,        
        n_hidden_features=5,
        lambda1=0.01,
        alpha1=0.5,
        lambda2=0.01,
        alpha2=0.5,
        family="gaussian",
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),  
        optimizer=Optimizer(),
        seed=123,
    ):

        super().__init__(
            n_hidden_features=n_hidden_features,
            lambda1=lambda1,
            alpha1=alpha1,
            lambda2=lambda2,
            alpha2=alpha2,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            optimizer=optimizer,
            seed=seed,
        )
        
        self.family = family
     
        
     
    def gaussian_loss(self, y, row_index, XB):
        return 0.5*np.mean(np.square(y[row_index] - XB))
        
    
    def laplace_loss(self, y, row_index, XB):
        return 0.5*np.mean(np.abs(y[row_index] - XB))

    
    def loglik_zip(beta, group_index, X, y, zero_mass, 
               lambda1, alpha1, lambda2, alpha2, 
               intercept=True):

        max_double = 709.0
        zero = np.finfo(float).eps
        not_zero_mass = 1 - zero_mass
        n = len(y)
    
        if intercept:
    
            XB = np.minimum(np.dot(np.column_stack((np.repeat(1, n), X)), beta), 
                            max_double) # (!) stack columns outside loglik
            
            mu = np.minimum(np.exp(XB), 
                          max_double) # avoid overflows  # but...
        
            exp_mu = np.exp(-mu)
            not_zero_mass_exp_mu = not_zero_mass*exp_mu
        
            f = (y == 0)*(zero_mass + not_zero_mass_exp_mu) \
            + (y > 0)*(not_zero_mass_exp_mu * np.power(mu, y)/factorial(y))   
        
            return -np.mean(np.log(np.maximum(f, zero)))
        
        else:
    
            XB = np.minimum(np.dot(X, beta), 
                              max_double) # avoid overflows  # but...
            
            mu = np.minimum(np.exp(XB), 
                          max_double) # avoid overflows # but...
        
            exp_mu = np.exp(-mu)
            not_zero_mass_exp_mu = not_zero_mass*exp_mu
        
            f = (y == 0)*(zero_mass + not_zero_mass_exp_mu) \
            + (y > 0)*(not_zero_mass_exp_mu * np.power(mu, y)/factorial(y))   
        
            return -np.mean(np.log(np.maximum(f, zero))) 
    

    def loglik_poisson(beta, group_index, X, y, 
                       lambda1, alpha1, lambda2, alpha2, 
                       intercept=True):
      
        max_double = 709.0
        zero = np.finfo(float).eps
        n = len(y)
    
        if intercept:
    
            XB = np.minimum(np.dot(np.column_stack((np.repeat(1, n), X)), beta), 
                            max_double) # (!) stack columns outside loglik
            
            mu = np.minimum(np.exp(XB), 
                          max_double) # avoid overflows # but...
        
            f = np.exp(-mu)*np.power(mu, y)/factorial(y)
        
            return -np.mean(np.log(np.maximum(f, zero)))
        
        else:
    
            XB = np.minimum(np.dot(X, beta), 
                              max_double) # avoid overflows # but...
            
            mu = np.minimum(np.exp(XB), 
                          max_double) # avoid overflows # but...
        
            f = np.exp(-mu)*np.power(mu, y)#/factorial(y)
        
            return -np.mean(np.log(np.maximum(f, zero)))
        
    
    def loglik_nbinom(beta, group_index, X, y, 
                      lambda1, alpha1, lambda2, alpha2, 
                      intercept=True):
      
      max_double = 709.0
      zero = np.finfo(float).eps
      n = len(y)
    
      if intercept:
        
        XB = np.minimum(np.dot(np.column_stack((np.repeat(1, n), X)), beta), 
                        max_double) # (!) stack columns outside loglik
    
        p_hat = 0.5*(1 + erf(1.0/(XB/n + 1.0)))
    
        f = factorial(y+n-1)/(factorial(n-1)*factorial(y))*np.power(p_hat, n)*np.power(1-p_hat, y) # nbinom.pmf(y, n, p_hat, loc=0)
    
        return -np.mean(np.log(np.maximum(f, zero)))
    
      else:
      
        XB = np.minimum(np.dot(X, beta), 
                        max_double) # avoid overflows
        
        p_hat = 0.5*(1 + erf(1.0/(XB/n + 1.0)))
    
        f = factorial(y+n-1)/(factorial(n-1)*factorial(y))*np.power(p_hat, n)*np.power(1-p_hat, y) # nbinom.pmf(y, n, p_hat, loc=0)
    
        return -np.mean(np.log(np.maximum(f, zero)))
    
    
    
    def loss_func(self, beta, group_index, X, y, 
                  row_index=None, type_loss="gaussian", 
                  **kwargs):

        res = {"gaussian": self.gaussian_loss,
               "laplace": self.laplace_loss}
        
        if row_index is None:            
            
            row_index = range(len(y))
            XB = self.compute_XB(X, beta=beta)  

            return res[type_loss](y, row_index, XB) + self.compute_penalty(group_index=group_index, 
                  beta=beta)              
            
        XB = self.compute_XB(X, beta=beta, 
                             row_index=row_index)                
        
        return res[type_loss](y, row_index, XB) + self.compute_penalty(group_index=group_index, 
                  beta=beta)                                          
        
                                                
    
    def fit(self, X, y, 
            learning_rate=0.01, decay=0.1, # in an object, in constructor
            batch_prop=1, tolerance=1e-5, # in an object, in constructor
            optimizer = None, # in an object, in constructor
            verbose=0, **kwargs):
        """Fit GLM model to training data (X, y).
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
            
            y: array-like, shape = [n_samples]
                Target values.
        
            **kwargs: additional parameters to be passed to 
                    self.cook_training_set or self.obj.fit
               
        Returns:
        
            self: object

        """
        
        self.beta = None
        
        self.n_iter = 0
        
        n, self.group_index = X.shape
        
        centered_y, scaled_Z = self.cook_training_set(y=y, X=X)
        
        n_Z = scaled_Z.shape[0]
        
        # initialization                    
        beta_ = np.linalg.lstsq(scaled_Z, centered_y, rcond=None)[0]     

        self.optimizer.learning_rate = learning_rate
        self.optimizer.decay = decay
        self.optimizer.batch_prop = batch_prop
        self.optimizer.verbose = verbose
         
        
        # optimization
        # fit(self, loss_func, response, x0, **kwargs):
        # loss_func(self, beta, group_index, X, y, 
        #          row_index=None, type_loss="gaussian", 
        #          **kwargs)
        self.optimizer.fit(self.loss_func,  
                           response = centered_y, 
                           x0 = beta_,
                           group_index = self.group_index, 
                           X = scaled_Z, 
                           y = centered_y, 
                           type_loss=self.family, 
                           tolerance=tolerance,
                           **kwargs)         

        self.beta = self.optimizer.results[0]
        
        return self
    
    
    
    def predict(self, X, **kwargs):
        """Predict test data X.
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
            
            **kwargs: additional parameters to be passed to 
                    self.cook_test_set
               
        Returns:
        
            model predictions: {array-like}        

        """
        
        if len(X.shape) == 1:

            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )
                           
            return (self.y_mean + np.dot(self.cook_test_set(new_X, **kwargs), 
                                         self.beta))[0]
        
        return self.y_mean + np.dot(self.cook_test_set(X, **kwargs), 
                                    self.beta)
    
    
    
    def score(self, X, y, scoring=None, **kwargs):
        """ Score the model on test set features X and response y. 

        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features

            y: array-like, shape = [n_samples]
                Target values

            scoring: str
                must be in ('explained_variance', 'neg_mean_absolute_error', \
                            'neg_mean_squared_error', 'neg_mean_squared_log_error', \
                            'neg_median_absolute_error', 'r2')
            
            **kwargs: additional parameters to be passed to scoring functions
               
        Returns: 
        
            model scores: {array-like}
            
        """

        preds = self.predict(X)

        if type(preds) == tuple:  # if there are std. devs in the predictions
            preds = preds[0]

        if scoring is None:
            scoring = "neg_mean_squared_error"

        # check inputs
        assert scoring in (
            "explained_variance",
            "neg_mean_absolute_error",
            "neg_mean_squared_error",
            "neg_mean_squared_log_error",
            "neg_median_absolute_error",
            "r2",
        ), "'scoring' should be in ('explained_variance', 'neg_mean_absolute_error', \
                           'neg_mean_squared_error', 'neg_mean_squared_log_error', \
                           'neg_median_absolute_error', 'r2')"

        scoring_options = {
            "explained_variance": skm2.explained_variance_score,
            "neg_mean_absolute_error": skm2.mean_absolute_error,
            "neg_mean_squared_error": skm2.mean_squared_error,
            "neg_mean_squared_log_error": skm2.mean_squared_log_error,
            "neg_median_absolute_error": skm2.median_absolute_error,
            "r2": skm2.r2_score,
        }

        return scoring_options[scoring](y, preds, **kwargs)


import numpy as np
import platform
from numpy import linalg as la
if platform.system() in ('Linux', 'Darwin'):
    import jax.numpy as jnp
    from jax.numpy import linalg as jla

# import .matrixops as mo
from . import matrixops as mo


# computes beta_hat = (t(x)%*%x + lam*I)^{-1}%*%t(x)%*%y
def beta_hat(x, y, lam=None, backend="cpu"):

    # assert on dimensions
    if lam is None:
        return mo.safe_sparse_dot(
            a=inv_penalized_cov(x=x, backend=backend),
            b=mo.crossprod(x=x, y=y, backend=backend),
            backend=backend,
        )

    return mo.safe_sparse_dot(
        a=inv_penalized_cov(x=x, lam=lam, backend=backend),
        b=mo.crossprod(x=x, y=y, backend=backend),
        backend=backend,
    )


# computes (t(x)%*%x + lam*I)^{-1}
def inv_penalized_cov(x, lam=None, backend="cpu"):
    # assert on dimensions
    sys_platform = platform.system()
    if backend in ("gpu", "tpu") and (sys_platform in ('Linux', 'Darwin')):
        if lam is None:
            return jla.inv(mo.crossprod(x=x, backend=backend))
        return jla.inv(
            mo.crossprod(x=x, backend=backend) + lam * jnp.eye(x.shape[1])
        )

    if lam is None:
        return la.inv(mo.crossprod(x))
    return la.inv(mo.crossprod(x) + lam * np.eye(x.shape[1]))


# linear regression with no regularization
def beta_Sigma_hat(
    X=None,
    y=None,
    fit_intercept=False,
    X_star=None,  # for preds
    return_cov=True,  # confidence interval required for preds?
    beta_hat_=None,  # for prediction only (X_star is not None)
    Sigma_hat_=None,
    backend="cpu",
):  # for prediction only (X_star is not None)

    if (X is not None) & (y is not None):  # fit

        if len(X.shape) == 1:
            X = X.reshape(-1, 1)

        if X_star is not None:
            if len(X_star.shape) == 1:
                X_star = X_star.reshape(-1, 1)

        n, p = X.shape

        if fit_intercept == True:
            X = mo.cbind(x=np.ones(n), y=X, backend=backend)
            if X_star is not None:
                X_star = mo.cbind(
                    x=np.ones(X_star.shape[0]), y=X_star, backend=backend
                )

        Cn = inv_penalized_cov(x=X, backend=backend)

        if return_cov == True:

            if X_star is None:

                beta_hat_ = mo.safe_sparse_dot(
                    a=Cn,
                    b=mo.crossprod(x=X, y=y, backend=backend),
                    backend=backend,
                )

                Sigma_hat_ = np.eye(X.shape[1]) - mo.safe_sparse_dot(
                    a=Cn, b=mo.crossprod(x=X, backend=backend), backend=backend
                )
                temp = mo.safe_sparse_dot(
                    a=Cn,
                    b=mo.tcrossprod(x=Sigma_hat_, y=X, backend=backend),
                    backend=backend,
                )
                smoothing_matrix = mo.safe_sparse_dot(
                    a=X, b=temp, backend=backend
                )
                y_hat = mo.safe_sparse_dot(
                    a=smoothing_matrix, b=y, backend=backend
                )

                return {
                    "beta_hat": beta_hat_,
                    "Sigma_hat": Sigma_hat_,
                    "GCV": np.mean(
                        ((y - y_hat) / (1 - np.trace(smoothing_matrix) / n))
                        ** 2
                    ),
                }

            else:

                if beta_hat_ is None:
                    beta_hat_ = mo.safe_sparse_dot(
                        a=Cn,
                        b=mo.crossprod(X, y, backend=backend),
                        backend=backend,
                    )

                if Sigma_hat_ is None:
                    Sigma_hat_ = np.eye(X.shape[1]) - mo.safe_sparse_dot(
                        a=Cn,
                        b=mo.crossprod(x=X, backend=backend),
                        backend=backend,
                    )

                temp = mo.safe_sparse_dot(
                    a=Cn,
                    b=mo.tcrossprod(Sigma_hat_, X, backend=backend),
                    backend=backend,
                )
                smoothing_matrix = mo.safe_sparse_dot(
                    a=X, b=temp, backend=backend
                )
                y_hat = mo.safe_sparse_dot(
                    a=smoothing_matrix, b=y, backend=backend
                )

                return {
                    "beta_hat": beta_hat_,
                    "Sigma_hat": Sigma_hat_,
                    "GCV": np.mean(
                        ((y - y_hat) / (1 - np.trace(smoothing_matrix) / n))
                        ** 2
                    ),
                    "preds": mo.safe_sparse_dot(
                        a=X_star, b=beta_hat_, backend=backend
                    ),
                    "preds_std": np.sqrt(
                        np.diag(
                            mo.safe_sparse_dot(
                                a=X_star,
                                b=mo.tcrossprod(
                                    x=Sigma_hat_, y=X_star, backend=backend
                                ),
                                backend=backend,
                            )
                        )
                    ),
                }

        else:  # return_cov == False

            if X_star is None:

                beta_hat_ = mo.safe_sparse_dot(
                    a=Cn,
                    b=mo.crossprod(x=X, y=y, backend=backend),
                    backend=backend,
                )
                Sigma_hat_ = np.eye(X.shape[1]) - mo.safe_sparse_dot(
                    a=Cn, b=mo.crossprod(x=X, backend=backend), backend=backend
                )
                temp = mo.safe_sparse_dot(
                    a=Cn,
                    b=mo.tcrossprod(x=Sigma_hat_, y=X, backend=backend),
                    backend=backend,
                )
                smoothing_matrix = mo.safe_sparse_dot(
                    a=X, b=temp, backend=backend
                )
                y_hat = mo.safe_sparse_dot(
                    a=smoothing_matrix, b=y, backend=backend
                )

                return {
                    "beta_hat": beta_hat_,
                    "GCV": np.mean(
                        ((y - y_hat) / (1 - np.trace(smoothing_matrix) / n))
                        ** 2
                    ),
                }

            else:

                if beta_hat_ is None:
                    beta_hat_ = beta_hat(X, y)
                    Sigma_hat_ = np.eye(X.shape[1]) - mo.safe_sparse_dot(
                        a=Cn,
                        b=mo.crossprod(x=X, backend=backend),
                        backend=backend,
                    )
                    temp = mo.safe_sparse_dot(
                        a=Cn,
                        b=mo.tcrossprod(x=Sigma_hat_, y=X, backend=backend),
                        backend=backend,
                    )
                    smoothing_matrix = mo.safe_sparse_dot(
                        a=X, b=temp, backend=backend
                    )
                    y_hat = mo.safe_sparse_dot(
                        a=smoothing_matrix, b=y, backend=backend
                    )

                return {
                    "beta_hat": beta_hat_,
                    "GCV": np.mean(
                        ((y - y_hat) / (1 - np.trace(smoothing_matrix) / n))
                        ** 2
                    ),
                    "preds": mo.safe_sparse_dot(
                        a=X_star, b=beta_hat_, backend=backend
                    ),
                }

    else:  # X is None | y is None  # predict

        assert (beta_hat_ is not None) & (X_star is not None)

        if return_cov == True:

            assert Sigma_hat_ is not None

            return {
                "preds": mo.safe_sparse_dot(
                    a=X_star, b=beta_hat_, backend=backend
                ),
                "preds_std": np.sqrt(
                    np.diag(
                        mo.safe_sparse_dot(
                            a=X_star,
                            b=mo.tcrossprod(
                                x=Sigma_hat_, y=X_star, backend=backend
                            ),
                            backend=backend,
                        )
                    )
                ),
            }

        else:

            return {
                "preds": mo.safe_sparse_dot(
                    a=X_star, b=beta_hat_, backend=backend
                )
            }


# beta and Sigma in Bayesian Ridge Regression 1
# without intercept! without intercept! without intercept!
def beta_Sigma_hat_rvfl(
    X=None,
    y=None,
    s=0.1,
    sigma=0.05,
    fit_intercept=False,
    X_star=None,  # for preds
    return_cov=True,  # confidence interval required for preds?
    beta_hat_=None,  # for prediction only (X_star is not None)
    Sigma_hat_=None,
    backend="cpu",
):  # for prediction only (X_star is not None)

    if (X is not None) & (y is not None):  # fit

        if len(X.shape) == 1:
            X = X.reshape(-1, 1)

        if X_star is not None:
            if len(X_star.shape) == 1:
                X_star = X_star.reshape(-1, 1)

        n, p = X.shape

        if fit_intercept == True:
            X = mo.cbind(np.ones(n), X, backend=backend)
            if X_star is not None:
                X_star = mo.cbind(
                    x=np.ones(X_star.shape[0]), y=X_star, backend=backend
                )

        s2 = s ** 2
        lambda_ = (sigma ** 2) / s2
        Cn = inv_penalized_cov(X, lam=lambda_)

        if return_cov == True:

            if X_star is None:

                beta_hat_ = mo.safe_sparse_dot(
                    a=Cn,
                    b=mo.crossprod(x=X, y=y, backend=backend),
                    backend=backend,
                )
                Sigma_hat_ = s2 * (
                    np.eye(X.shape[1])
                    - mo.safe_sparse_dot(
                        a=Cn,
                        b=mo.crossprod(x=X, backend=backend),
                        backend=backend,
                    )
                )
                temp = mo.safe_sparse_dot(
                    a=Cn,
                    b=mo.tcrossprod(x=Sigma_hat_, y=X, backend=backend),
                    backend=backend,
                )
                smoothing_matrix = mo.safe_sparse_dot(
                    a=X, b=temp, backend=backend
                )
                y_hat = mo.safe_sparse_dot(
                    a=smoothing_matrix, b=y, backend=backend
                )

                return {
                    "beta_hat": beta_hat_,
                    "Sigma_hat": Sigma_hat_,
                    "GCV": np.mean(
                        ((y - y_hat) / (1 - np.trace(smoothing_matrix) / n))
                        ** 2
                    ),
                }

            else:

                if beta_hat_ is None:
                    beta_hat_ = mo.safe_sparse_dot(
                        a=Cn,
                        b=mo.crossprod(x=X, y=y, backend=backend),
                        backend=backend,
                    )

                if Sigma_hat_ is None:
                    Sigma_hat_ = s2 * (
                        np.eye(X.shape[1])
                        - mo.safe_sparse_dot(
                            a=Cn,
                            b=mo.crossprod(x=X, backend=backend),
                            backend=backend,
                        )
                    )

                temp = mo.safe_sparse_dot(
                    a=Cn,
                    b=mo.tcrossprod(x=Sigma_hat_, y=X, backend=backend),
                    backend=backend,
                )
                smoothing_matrix = mo.safe_sparse_dot(
                    a=X, b=temp, backend=backend
                )
                y_hat = mo.safe_sparse_dot(
                    a=smoothing_matrix, b=y, backend=backend
                )

                return {
                    "beta_hat": beta_hat_,
                    "Sigma_hat": Sigma_hat_,
                    "GCV": np.mean(
                        ((y - y_hat) / (1 - np.trace(smoothing_matrix) / n))
                        ** 2
                    ),
                    "preds": mo.safe_sparse_dot(
                        a=X_star, b=beta_hat_, backend=backend
                    ),
                    "preds_std": np.sqrt(
                        np.diag(
                            mo.safe_sparse_dot(
                                a=X_star,
                                b=mo.tcrossprod(
                                    x=Sigma_hat_, y=X_star, backend=backend
                                ),
                                backend=backend,
                            )
                            + (sigma ** 2) * np.eye(X_star.shape[0])
                        )
                    ),
                }

        else:  # return_cov == False

            if X_star is None:

                beta_hat_ = mo.safe_sparse_dot(
                    a=Cn,
                    b=mo.crossprod(x=X, y=y, backend=backend),
                    backend=backend,
                )
                Sigma_hat_ = s2 * (
                    np.eye(X.shape[1])
                    - mo.safe_sparse_dot(
                        a=Cn, b=mo.crossprod(X), backend=backend
                    )
                )
                temp = mo.safe_sparse_dot(
                    a=Cn,
                    b=mo.tcrossprod(x=Sigma_hat_, y=X, backend=backend),
                    backend=backend,
                )
                smoothing_matrix = mo.safe_sparse_dot(
                    a=X, b=temp, backend=backend
                )
                y_hat = mo.safe_sparse_dot(
                    a=smoothing_matrix, b=y, backend=backend
                )

                return {
                    "beta_hat": beta_hat_,
                    "GCV": np.mean(
                        ((y - y_hat) / (1 - np.trace(smoothing_matrix) / n))
                        ** 2
                    ),
                }

            else:

                if beta_hat_ is None:
                    beta_hat_ = beta_hat(X, y, lam=lambda_, backend=backend)
                    Sigma_hat_ = s2 * (
                        np.eye(X.shape[1])
                        - mo.safe_sparse_dot(
                            a=Cn,
                            b=mo.crossprod(x=X, backend=backend),
                            backend=backend,
                        )
                    )
                    temp = mo.safe_sparse_dot(
                        a=Cn,
                        b=mo.tcrossprod(x=Sigma_hat_, y=X, backend=backend),
                        backend=backend,
                    )
                    smoothing_matrix = mo.safe_sparse_dot(
                        a=X, b=temp, backend=backend
                    )
                    y_hat = mo.safe_sparse_dot(
                        a=smoothing_matrix, b=y, backend=backend
                    )

                return {
                    "beta_hat": beta_hat_,
                    "GCV": np.mean(
                        ((y - y_hat) / (1 - np.trace(smoothing_matrix) / n))
                        ** 2
                    ),
                    "preds": mo.safe_sparse_dot(
                        a=X_star, b=beta_hat_, backend=backend
                    ),
                }

    else:  # X is None | y is None  # predict

        assert (beta_hat_ is not None) & (X_star is not None)

        if return_cov == True:

            assert Sigma_hat_ is not None

            return {
                "preds": mo.safe_sparse_dot(X_star, beta_hat_, backend=backend),
                "preds_std": np.sqrt(
                    np.diag(
                        mo.safe_sparse_dot(
                            a=X_star,
                            b=mo.tcrossprod(
                                x=Sigma_hat_, y=X_star, backend=backend
                            ),
                            backend=backend,
                        )
                        + (sigma ** 2) * np.eye(X_star.shape[0])
                    )
                ),
            }

        else:

            return {
                "preds": mo.safe_sparse_dot(
                    a=X_star, b=beta_hat_, backend=backend
                )
            }


# beta and Sigma in Bayesian Ridge Regression 2
# without intercept! without intercept! without intercept!
def beta_Sigma_hat_rvfl2(
    X=None,
    y=None,
    Sigma=None,
    sigma=0.05,
    fit_intercept=False,
    X_star=None,  # check when dim = 1 # check when dim = 1
    return_cov=True,
    beta_hat_=None,  # for prediction only (X_star is not None)
    Sigma_hat_=None,
    backend="cpu",
):  # for prediction only (X_star is not None)

    if (X is not None) & (y is not None):

        if len(X.shape) == 1:
            X = X.reshape(-1, 1)

        n, p = X.shape

        if Sigma is None:
            if fit_intercept == True:
                Sigma = np.eye(p + 1)
            else:
                Sigma = np.eye(p)

        if X_star is not None:
            if len(X_star.shape) == 1:
                X_star = X_star.reshape(-1, 1)

        if fit_intercept == True:

            X = mo.cbind(np.ones(n), X)
            Cn = (
                la.inv(
                    mo.safe_sparse_dot(
                        a=Sigma,
                        b=mo.crossprod(x=X, backend=backend),
                        backend="cpu",
                    )
                    + (sigma ** 2) * np.eye(p + 1)
                )
                if backend == "cpu"
                else jla.inv(
                    mo.safe_sparse_dot(
                        a=Sigma,
                        b=mo.crossprod(x=X, backend=backend),
                        backend=backend,
                    )
                    + (sigma ** 2) * np.eye(p + 1)
                )
            )

            if X_star is not None:
                X_star = mo.cbind(
                    x=np.ones(X_star.shape[0]), y=X_star, backend=backend
                )
        else:
            # rename to invCn
            Cn = (
                la.inv(
                    mo.safe_sparse_dot(
                        a=Sigma,
                        b=mo.crossprod(x=X, backend=backend),
                        backend="cpu",
                    )
                    + (sigma ** 2) * np.eye(p)
                )
                if backend == "cpu"
                else jla.inv(
                    mo.safe_sparse_dot(
                        a=Sigma,
                        b=mo.crossprod(x=X, backend=backend),
                        backend=backend,
                    )
                    + (sigma ** 2) * np.eye(p)
                )
            )

        temp = mo.safe_sparse_dot(
            a=Cn,
            b=mo.tcrossprod(x=Sigma, y=X, backend=backend),
            backend=backend,
        )
        smoothing_matrix = mo.safe_sparse_dot(a=X, b=temp, backend=backend)
        y_hat = mo.safe_sparse_dot(a=smoothing_matrix, b=y, backend=backend)

        if return_cov == True:

            if X_star is None:

                return {
                    "beta_hat": mo.safe_sparse_dot(
                        a=temp, b=y, backend=backend
                    ),
                    "Sigma_hat": Sigma
                    - mo.safe_sparse_dot(
                        a=temp,
                        b=mo.safe_sparse_dot(a=X, b=Sigma, backend=backend),
                        backend=backend,
                    ),
                    "GCV": np.mean(
                        ((y - y_hat) / (1 - np.trace(smoothing_matrix) / n))
                        ** 2
                    ),
                }

            else:

                if beta_hat_ is None:
                    beta_hat_ = mo.safe_sparse_dot(a=temp, b=y, backend=backend)

                if Sigma_hat_ is None:
                    Sigma_hat_ = Sigma - mo.safe_sparse_dot(
                        a=temp,
                        b=mo.safe_sparse_dot(a=X, b=Sigma, backend=backend),
                        backend=backend,
                    )

                return {
                    "beta_hat": beta_hat_,
                    "Sigma_hat": Sigma_hat_,
                    "GCV": np.mean(
                        ((y - y_hat) / (1 - np.trace(smoothing_matrix) / n))
                        ** 2
                    ),
                    "preds": mo.safe_sparse_dot(
                        a=X_star, b=beta_hat_, backend=backend
                    ),
                    "preds_std": np.sqrt(
                        np.diag(
                            mo.safe_sparse_dot(
                                a=X_star,
                                b=mo.tcrossprod(
                                    x=Sigma_hat_, y=X_star, backend=backend
                                ),
                                backend=backend,
                            )
                            + (sigma ** 2) * np.eye(X_star.shape[0])
                        )
                    ),
                }

        else:  # return_cov == False

            if X_star is None:

                return {
                    "beta_hat": mo.safe_sparse_dot(
                        a=temp, b=y, backend=backend
                    ),
                    "GCV": np.mean(
                        ((y - y_hat) / (1 - np.trace(smoothing_matrix) / n))
                        ** 2
                    ),
                }

            else:

                if beta_hat_ is None:
                    beta_hat_ = mo.safe_sparse_dot(a=temp, b=y, backend=backend)

                return {
                    "beta_hat": beta_hat_,
                    "GCV": np.mean(
                        ((y - y_hat) / (1 - np.trace(smoothing_matrix) / n))
                        ** 2
                    ),
                    "preds": mo.safe_sparse_dot(
                        a=X_star, b=beta_hat_, backend=backend
                    ),
                }

    else:  # (X is None) | (y is None) # predict

        assert (beta_hat_ is not None) & (X_star is not None)

        if return_cov == True:

            assert Sigma_hat_ is not None

            return {
                "preds": mo.safe_sparse_dot(
                    a=X_star, b=beta_hat_, backend=backend
                ),
                "preds_std": np.sqrt(
                    np.diag(
                        mo.safe_sparse_dot(
                            a=X_star,
                            b=mo.tcrossprod(
                                Sigma_hat_, X_star, backend=backend
                            ),
                            backend=backend,
                        )
                        + (sigma ** 2) * np.eye(X_star.shape[0])
                    )
                ),
            }

        else:

            return {
                "preds": mo.safe_sparse_dot(
                    a=X_star, b=beta_hat_, backend=backend
                )
            }



# cython: wraparound=False
# cython: boundscheck=False
# cython: nonecheck=False

# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
cimport numpy as np
cimport cython
from collections import Counter


# 0 - utils -----

# 0 - 0 data structures & funcs -----

ctypedef fused nparray_double:    
    double

cdef list flatten(x): 
            
    cdef list res = []    
    
    for elt1 in x:
                    
        try:
            
            for elt2 in elt1:
                
                res.append(elt2)  
                
        except:
            
            res.append(elt1)  
                                
    return res


def is_factor(y):

    cdef long int n = len(y)
    cdef long int idx = 0
    ans = True    
    
    # check if x is int
    def is_int(x):
        try:
            return int(x) == x
        except:
            return False

    # check if x is float
    def is_float(x):
        return isinstance(x, float)
    
    for idx in range(n):
        
        if is_int(y[idx]) & (is_float(y[idx]) == False):
            
            idx += 1
            
        else:
            
            ans = False
            
            break

    return ans


# 1 - stratified subsampling based on the response -----

def subsamplec(y, double row_sample=0.8, int seed=123):
    
    cdef long int i, n_classes, n_obs, n_obs_out
    cdef long int[:] n_elem_classes
    cdef double[:] freqs_hist
    cdef list index = []

    assert (row_sample < 1) & (
        row_sample >= 0
    ), "'row_sample' must be < 1 and >= 0"

    n_obs = len(y)
    n_obs_out = np.ceil(n_obs * row_sample)

    # preproc -----
    if is_factor(y):

        classes, n_elem_classes = np.unique(y, return_counts=True)
        n_classes = len(classes)
        y_as_classes = y.copy()
        freqs_hist = np.zeros_like(n_elem_classes, dtype=np.float)
        
        for i in range(len(n_elem_classes)):
            freqs_hist[i] = np.float(n_elem_classes[i]) / n_obs

    else:

        h = np.histogram(y, bins="auto")
        n_elem_classes = np.asarray(h[0], dtype=np.int)
        freqs_hist = np.zeros_like(n_elem_classes, dtype=np.float)
        
        for i in range(len(n_elem_classes)):
            freqs_hist[i] = np.float(n_elem_classes[i]) / n_obs
        
        breaks = h[1]

        n_breaks_1 = len(breaks) - 1
        classes = range(n_breaks_1)
        n_classes = n_breaks_1
        y_as_classes = np.zeros_like(y, dtype=int)

        for i in classes:
            y_as_classes[(y > breaks[i]) * (y <= breaks[i + 1])] = int(i)

    # main loop ----

    np.random.seed(seed)

    for i in range(n_classes):

        bool_class_i = (y_as_classes == classes[i])               

        # index_class_i = [i for i, e in enumerate(bool_class_i) if e == True]
        index_class_i = np.asarray(np.where(bool_class_i == True)[0], dtype=np.int)

        if np.sum(bool_class_i) > 1:  # at least 2 elements in class  #i

            index.append(
                np.random.choice(
                    index_class_i,
                    size=int(n_obs_out * freqs_hist[i]),  # output size
                    replace=True,
                ).tolist()
            )

        else:  # only one element in class

            try:

                index.append(index_class_i[0])

            except:

                0
    
    return np.asarray(flatten(index))
    


# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
from ..base import Base
from scipy.stats import norm
from ..utils import matrixops as mo
from ..utils import timeseries as ts
import sklearn.metrics as skm2
import pickle
from functools import partial


class MTS(Base):
    """Univariate and multivariate time series (MTS) forecasting with Quasi-Randomized networks
    
    Attributes: 
     
        obj: object.
            any object containing a method fit (obj.fit()) and a method predict 
            (obj.predict()).

        n_hidden_features: int.
            number of nodes in the hidden layer.

        activation_name: str.
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'.

        a: float.
            hyperparameter for 'prelu' or 'elu' activation function.

        nodes_sim: str.
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
            'uniform'.

        bias: boolean.
            indicates if the hidden layer contains a bias term (True) or not 
            (False).

        dropout: float.
            regularization parameter; (random) percentage of nodes dropped out 
            of the training.

        direct_link: boolean.
            indicates if the original predictors are included (True) in model's fitting or not (False).

        n_clusters: int.
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0: no clustering).

        cluster_encode: bool.
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding.

        type_clust: str.
            type of clustering method: currently k-means ('kmeans') or Gaussian 
            Mixture Model ('gmm').

        type_scaling: a tuple of 3 strings.
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant). 
            Currently available: standardization ('std') or MinMax scaling ('minmax').

        col_sample: float.
            percentage of covariates randomly chosen for training. 

        lags: int.
            number of lags used for each time series. 

        alpha1: float.
            smoothing weight (training set).

        alpha2: float.
            smoothing parameter (testing set). 

        seed: int. 
            reproducibility seed for nodes_sim=='uniform'.
            
        backend: str.
            "cpu" or "gpu" or "tpu".                           
    """

    # construct the object -----

    def __init__(
        self,
        obj,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        lags=1,
        alpha1=None,
        alpha2=None,
        seed=123,
        backend="cpu"
    ):

        assert np.int(lags) == lags, "parameter 'lags' should be an integer"

        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            seed=seed,
            backend=backend
        )

        self.obj = obj
        self.n_series = None
        self.lags = lags
        self.fit_objs = {}
        self.y = None  # MTS responses (most recent observations first)
        self.X = None  # MTS lags
        self.xreg = None
        self.y_means = {}
        self.preds = None
        self.preds_std = []
        self.row_sample = 1
        self.alpha1 = alpha1
        self.alpha2 = alpha2

    def fit(self, X, xreg=None, **kwargs):
        """Fit MTS model to training data X, with optional regressors xreg
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training time series, where n_samples is the number 
                of samples and n_features is the number of features;
                X must be in increasing order (most recent observations last)

            xreg: {array-like}, shape = [n_samples, n_features_xreg]
                Additional regressors to be passed to obj
                xreg must be in increasing order (most recent observations last)

            **kwargs: additional parameters to be passed to 
                    self.cook_training_set
               
        Returns: 
        
            self: object
        """

        try:
            # multivariate time series
            n, p = X.shape
        except:
            # univariate time series
            n = X.shape[0]
            p = 1

        rep_1_n = np.repeat(1, n)

        self.y = None
        self.X = None
        self.n_series = p
        self.fit_objs.clear()
        self.y_means.clear()

        if p > 1:
            # multivariate time series
            mts_input = ts.create_train_inputs(X[::-1], self.lags)
        else:
            # univariate time series
            mts_input = ts.create_train_inputs(
                X.reshape(-1, 1)[::-1], self.lags
            )

        self.y = mts_input[0]
        # print(f"self.y: \n {self.y} \n ")
        # print(f"mts_input[1]: \n {mts_input[1]} \n ")
        if self.alpha1 is not None:
            alphas_seq = np.asarray(
                [
                    self.alpha1 * ((1 - self.alpha1) ** i)
                    for i in range(n - 1, -1, -1)
                ]
            )  # first to last
            alphas_seq_lags = ts.create_lags(
                alphas_seq, self.lags - 1, self.n_series
            )[:-1, :]
            # print(f"alphas_seq_lags: \n {alphas_seq_lags}")
            self.X = mts_input[1] * np.exp(alphas_seq_lags)
        else:
            self.X = mts_input[1]

        if xreg is not None:

            assert (
                xreg.shape[0] == n
            ), "'xreg' and 'X' must have the same number or observations"

            self.xreg = xreg

            xreg_input = ts.create_train_inputs(xreg[::-1], self.lags)

            dummy_y, scaled_Z = self.cook_training_set(
                y=rep_1_n,
                X=mo.cbind(self.X, xreg_input[1], backend=self.backend),
                **kwargs
            )

        else:  # xreg is None

            # avoids scaling X p times in the loop
            dummy_y, scaled_Z = self.cook_training_set(
                y=rep_1_n, X=self.X, **kwargs
            )

        # loop on all the time series and adjust self.obj.fit
        for i in range(p):
            y_mean = np.mean(self.y[:, i])
            self.y_means[i] = y_mean
            self.fit_objs[i] = pickle.loads(
                pickle.dumps(
                    self.obj.fit(scaled_Z, self.y[:, i] - y_mean, **kwargs), -1
                )
            )

        return self

    def predict(self, h=5, level=95, new_xreg=None, **kwargs):
        """Forecast all the time series, h steps ahead
        
        Args:
        
            h: {integer}
                Forecasting horizon
            
            level: {integer}
                Level of confidence (if obj has option 'return_std' and the 
                posterior is gaussian)
                
            new_xreg: {array-like}, shape = [n_samples = h, n_new_xreg]
                New values of additional (deterministic) regressors on horizon = h
                new_xreg must be in increasing order (most recent observations last)
            
            **kwargs: additional parameters to be passed to 
                    self.cook_test_set
               
        Returns: 
        
            model predictions for horizon = h: {array-like}
        """

        self.return_std = False

        if self.xreg is not None:
            assert new_xreg is not None, "'new_xreg' must be provided"

        self.preds = None  # do not remove (!)

        self.preds = pickle.loads(pickle.dumps(self.y, -1))

        if self.alpha2 is not None:
            if len(self.preds) > 1:
                # multivariate
                previous_pred = self.preds[0, :]
            else:
                # univariate
                previous_pred = self.preds[0]

        n_features = self.n_series * self.lags

        if "return_std" in kwargs:

            self.preds_std = np.zeros(h)

            self.return_std = kwargs["return_std"]

            qt = norm.ppf(1 - 0.5 * (1 - level / 100))

        if new_xreg is not None:  # Additional regressors provided

            try:
                n_obs_xreg, n_features_xreg = new_xreg.shape
                assert (
                    n_features_xreg == self.xreg.shape[1]
                ), "check number of series provided for 'new_xreg' (compare with self.xreg.shape[1])"
            except:
                n_obs_xreg = new_xreg.shape  # one series

            assert (
                n_obs_xreg == h
            ), "please provide values of regressors 'new_xreg' for the whole horizon 'h'"

            n_features_xreg = n_features_xreg * self.lags

            n_features_total = n_features + n_features_xreg

            inv_new_xreg = mo.rbind(self.xreg, new_xreg)[::-1]

            # Loop on horizon h
            for i in range(h):

                new_obs = ts.reformat_response(self.preds, self.lags)

                new_obs_xreg = ts.reformat_response(inv_new_xreg, self.lags)

                new_X = mo.rbind(
                    np.union1d(
                        new_obs.reshape(1, n_features),
                        new_obs_xreg.reshape(1, n_features_xreg),
                    ),
                    np.ones(n_features_total).reshape(1, n_features_total),
                )

                cooked_new_X = self.cook_test_set(new_X, **kwargs)

                predicted_cooked_new_X = self.obj.predict(
                    cooked_new_X, **kwargs
                )

                if self.return_std == False:  # std. dev. is not returned

                    if self.alpha2 is not None:

                        preds = (
                            self.alpha2
                            * np.array(
                                [
                                    (
                                        self.y_means[j]
                                        + predicted_cooked_new_X[0]
                                    )
                                    for j in range(self.n_series)
                                ]
                            )
                            + (1 - self.alpha2) * previous_pred
                        )

                    else:

                        preds = np.array(
                            [
                                (self.y_means[j] + predicted_cooked_new_X[0])
                                for j in range(self.n_series)
                            ]
                        )

                else:  # std. dev. is returned

                    if self.alpha2 is not None:

                        preds = (
                            self.alpha2
                            * np.array(
                                [
                                    (
                                        self.y_means[j]
                                        + predicted_cooked_new_X[0][0]
                                    )
                                    for j in range(self.n_series)
                                ]
                            )
                            + (1 - self.alpha2) * previous_pred
                        )

                    else:

                        preds = np.array(
                            [
                                (self.y_means[j] + predicted_cooked_new_X[0][0])
                                for j in range(self.n_series)
                            ]
                        )

                    self.preds_std[i] = predicted_cooked_new_X[1][0]

                if self.alpha2 is not None:

                    previous_pred = pickle.loads(pickle.dumps(preds, -1))

                self.preds = mo.rbind(preds, self.preds)

        else:  # No additional regressor provided

            for i in range(h):

                new_obs = ts.reformat_response(self.preds, self.lags)

                new_X = mo.rbind(
                    new_obs.reshape(1, n_features),
                    np.ones(n_features).reshape(1, n_features),
                )

                cooked_new_X = self.cook_test_set(new_X, **kwargs)

                predicted_cooked_new_X = self.obj.predict(
                    cooked_new_X, **kwargs
                )

                if self.return_std == False:  # std. dev. is not returned

                    if self.alpha2 is not None:

                        preds = (
                            self.alpha2
                            * np.array(
                                [
                                    (
                                        self.y_means[j]
                                        + predicted_cooked_new_X[0]
                                    )
                                    for j in range(self.n_series)
                                ]
                            )
                            + (1 - self.alpha2) * previous_pred
                        )

                    else:

                        preds = np.array(
                            [
                                (self.y_means[j] + predicted_cooked_new_X[0])
                                for j in range(self.n_series)
                            ]
                        )

                else:  # std. dev. is returned

                    if self.alpha2 is not None:

                        preds = (
                            self.alpha2
                            * np.array(
                                [
                                    (
                                        self.y_means[j]
                                        + predicted_cooked_new_X[0][0]
                                    )
                                    for j in range(self.n_series)
                                ]
                            )
                            + (1 - self.alpha2) * previous_pred
                        )

                    else:

                        preds = np.array(
                            [
                                (self.y_means[j] + predicted_cooked_new_X[0][0])
                                for j in range(self.n_series)
                            ]
                        )

                    self.preds_std[i] = predicted_cooked_new_X[1][0]

                if self.alpha2 is not None:
                    previous_pred = pickle.loads(pickle.dumps(preds, -1))

                self.preds = mo.rbind(preds, self.preds)

        # function's return

        self.preds = self.preds[0:h, :][::-1]

        if self.return_std == False:  # std. dev. is returned

            return self.preds

        # std. dev. is not returned
        self.preds_std = self.preds_std[::-1].reshape(h, 1)

        self.preds_std = np.repeat(self.preds_std, self.n_series).reshape(
            -1, self.n_series
        )

        return (
            self.preds,
            self.preds_std,
            self.preds - qt * self.preds_std,
            self.preds + qt * self.preds_std,
        )

    def score(self, X, training_index, testing_index, scoring=None, **kwargs):
        """ Train on training_index, score on testing_index. """

        assert (
            bool(set(training_index).intersection(set(testing_index))) == False
        ), "Non-overlapping 'training_index' and 'testing_index' required"

        # Dimensions
        n, p = X.shape

        # Training and testing sets
        X_train = X[training_index, :]
        X_test = X[testing_index, :]

        # Horizon
        h = len(testing_index)
        assert (
            len(training_index) + h
        ) <= n, "Please check lengths of training and testing windows"

        # Fit and predict
        self.fit(X_train, **kwargs)
        preds = self.predict(h=h, **kwargs)

        if scoring is None:
            scoring = "neg_root_mean_squared_error"

        # check inputs
        assert scoring in (
            "explained_variance",
            "neg_mean_absolute_error",
            "neg_mean_squared_error",
            "neg_root_mean_squared_error",
            "neg_mean_squared_log_error",
            "neg_median_absolute_error",
            "r2",
        ), "'scoring' should be in ('explained_variance', 'neg_mean_absolute_error', \
                               'neg_mean_squared_error', 'neg_root_mean_squared_error', 'neg_mean_squared_log_error', \
                               'neg_median_absolute_error', 'r2')"

        scoring_options = {
            "explained_variance": skm2.explained_variance_score,
            "neg_mean_absolute_error": skm2.mean_absolute_error,
            "neg_mean_squared_error": skm2.mean_squared_error,
            "neg_root_mean_squared_error": lambda x, y: np.sqrt(
                skm2.mean_squared_error(x, y)
            ),
            "neg_mean_squared_log_error": skm2.mean_squared_log_error,
            "neg_median_absolute_error": skm2.median_absolute_error,
            "r2": skm2.r2_score,
        }

        return tuple(
            [
                scoring_options[scoring](X_test[:, i], preds[:, i], **kwargs)
                for i in range(p)
            ]
        )



from sys import version_info as _swig_python_version_info

if _swig_python_version_info < (2, 7, 0):
    raise RuntimeError("Python 2.7 or later required")

    # Import the low-level C/C++ module
    # if __package__ or "." in __name__:

    try:

        from ._halton import *

    except:

        import _halton

    else:

        try:
            import builtins as __builtin__
        except ImportError:
            import __builtin__

        def _swig_repr(self):
            try:
                strthis = "proxy of " + self.this.__repr__()
            except __builtin__.Exception:
                strthis = ""
            return "<%s.%s; %s >" % (
                self.__class__.__module__,
                self.__class__.__name__,
                strthis,
            )

        def _swig_setattr_nondynamic_instance_variable(set):
            def set_instance_attr(self, name, value):
                if name == "thisown":
                    self.this.own(value)
                elif name == "this":
                    set(self, name, value)
                elif hasattr(self, name) and isinstance(
                    getattr(type(self), name), property
                ):
                    set(self, name, value)
                else:
                    raise AttributeError(
                        "You cannot add instance attributes to %s" % self
                    )

            return set_instance_attr

        def _swig_setattr_nondynamic_class_variable(set):
            def set_class_attr(cls, name, value):
                if hasattr(cls, name) and not isinstance(
                    getattr(cls, name), property
                ):
                    set(cls, name, value)
                else:
                    raise AttributeError(
                        "You cannot add class attributes to %s" % cls
                    )

            return set_class_attr

        def _swig_add_metaclass(metaclass):
            """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""

            def wrapper(cls):
                return metaclass(
                    cls.__name__, cls.__bases__, cls.__dict__.copy()
                )

            return wrapper

        class _SwigNonDynamicMeta(type):
            """Meta class to enforce nondynamic attributes (no new attributes) for a class"""

            __setattr__ = _swig_setattr_nondynamic_class_variable(
                type.__setattr__
            )

        def halton(i, m):
            return _halton.halton(i, m)

        def halton_base(i, m, b):
            return _halton.halton_base(i, m, b)

        def halton_inverse(r, m):
            return _halton.halton_inverse(r, m)

        def halton_sequence(i1, i2, m):
            return _halton.halton_sequence(i1, i2, m)

        def i4vec_sum(n, a):
            return _halton.i4vec_sum(n, a)

        def prime(n):
            return _halton.prime(n)

        def r8_mod(x, y):
            return _halton.r8_mod(x, y)

        def timestamp():
            return _halton.timestamp()




try:

    from .nodesimulation import (
        generate_sobol2,
        generate_sobol_randtoolbox,
        generate_uniform,
        generate_hammersley,
        generate_halton,
        generate_halton_randtoolbox,
    )

    __all__ = [
        "generate_sobol2",
        "generate_sobol_randtoolbox",
        "generate_uniform",
        "generate_hammersley",
        "generate_halton",
        "generate_halton_randtoolbox"
    ]

except:

    from .nodesimulation import (
        generate_sobol2,
        generate_uniform,
        generate_hammersley,
        generate_halton,
    )

    __all__ = [
        "generate_sobol2",
        "generate_uniform",
        "generate_hammersley",
        "generate_halton"
    ]



# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
import sklearn.metrics as skm2
from .bag import RandomBag
from ..custom import CustomClassifier
from ..utils import misc as mx
from ..utils import Progbar
from sklearn.base import ClassifierMixin
import pickle
from joblib import Parallel, delayed
from tqdm import tqdm


class RandomBagClassifier(RandomBag, ClassifierMixin):
    """Randomized 'Bagging' Classification model 
    
    Attributes: 
    
        obj: object
            any object containing a method fit (obj.fit()) and a method predict 
            (obj.predict())

        n_estimators: int
            number of boosting iterations

        n_hidden_features: int
            number of nodes in the hidden layer

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not 
            (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out 
            of the training

        direct_link: boolean
            indicates if the original predictors are included (True) in model's 
            fitting or not (False)

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0: 
                no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian 
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant). 
            Currently available: standardization ('std') or MinMax scaling ('minmax')

        col_sample: float
            percentage of covariates randomly chosen for training   

        row_sample: float
            percentage of rows chosen for training, by stratified bootstrapping 

        seed: int 
            reproducibility seed for nodes_sim=='uniform'

        backend: str
            "cpu" or "gpu" or "tpu"                           

    """

    # construct the object -----

    def __init__(
        self,
        obj,
        n_estimators=10,
        n_hidden_features=1,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=False,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        n_jobs=None,
        seed=123,
        verbose=1,
        backend="cpu"
    ):

        super().__init__(
            obj=obj,
            n_estimators=n_estimators,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend
        )

        self.type_fit = "classification"
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.voter = dict.fromkeys(range(n_estimators))

    def fit(self, X, y, **kwargs):
        """Fit Random 'Forest' model to training data (X, y).
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
            
            y: array-like, shape = [n_samples]
                Target values.
        
            **kwargs: additional parameters to be passed to 
                    self.cook_training_set or self.obj.fit
                
        Returns: 
        
            self: object
            
        """

        assert mx.is_factor(y), "y must contain only integers"

        # training
        n, p = X.shape
        self.n_classes = len(np.unique(y))

        base_learner = CustomClassifier(
            self.obj,
            n_hidden_features=self.n_hidden_features,
            activation_name=self.activation_name,
            a=self.a,
            nodes_sim=self.nodes_sim,
            bias=self.bias,
            dropout=self.dropout,
            direct_link=self.direct_link,
            n_clusters=self.n_clusters,
            type_clust=self.type_clust,
            type_scaling=self.type_scaling,
            col_sample=self.col_sample,
            row_sample=self.row_sample,
            seed=self.seed,
        )

        if self.verbose == 1:
            pbar = Progbar(self.n_estimators)

        # 1 - Sequential training -----

        if self.n_jobs is None:

            for idx, m in enumerate(range(self.n_estimators)):

                try:

                    base_learner.fit(X, y, **kwargs)

                    self.voter.update(
                        {m: pickle.loads(pickle.dumps(base_learner, -1))}
                    )

                    base_learner.set_params(seed=self.seed + (m + 1) * 1000)

                    if self.verbose == 1:
                        pbar.update(m)

                except:

                    if self.verbose == 1:
                        pbar.update(m)

                    continue

            if self.verbose == 1:
                pbar.update(self.n_estimators)

            self.n_estimators = len(self.voter)

            return self

        # 2 - Parallel training -----

        # if self.n_jobs is not None:
        def fit_estimators(m):
            try:
                base_learner = CustomClassifier(
                    self.obj,
                    n_hidden_features=self.n_hidden_features,
                    activation_name=self.activation_name,
                    a=self.a,
                    nodes_sim=self.nodes_sim,
                    bias=self.bias,
                    dropout=self.dropout,
                    direct_link=self.direct_link,
                    n_clusters=self.n_clusters,
                    type_clust=self.type_clust,
                    type_scaling=self.type_scaling,
                    col_sample=self.col_sample,
                    row_sample=self.row_sample,
                    seed=self.seed + m * 1000,
                )
                base_learner.fit(X, y, **kwargs)
                self.voter.update(
                    {m: pickle.loads(pickle.dumps(base_learner, -1))}
                )
            except:
                pass

        if self.verbose == 1:

            Parallel(n_jobs=self.n_jobs, prefer="threads")(
                delayed(fit_estimators)(m)
                for m in tqdm(range(self.n_estimators))
            )
        else:

            Parallel(n_jobs=self.n_jobs, prefer="threads")(
                delayed(fit_estimators)(m) for m in range(self.n_estimators)
            )

        self.n_estimators = len(self.voter)

        return self

    def predict(self, X, weights=None, **kwargs):
        """Predict test data X.
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
            
            **kwargs: additional parameters to be passed to 
                    self.cook_test_set
               
        Returns: 
        
            model predictions: {array-like}        

        """
        return self.predict_proba(X, weights, **kwargs).argmax(axis=1)

    def predict_proba(self, X, weights=None, **kwargs):
        """Predict probabilities for test data X.
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.
            
            **kwargs: additional parameters to be passed to 
                    self.cook_test_set
               
        Returns: 
        
            probability estimates for test data: {array-like}  

        """

        def calculate_probas(voter, weights=None, verbose=None):

            ensemble_proba = 0

            n_iter = len(voter)

            assert n_iter > 0, "no estimator found in `RandomBag` ensemble"

            if weights is None:

                for idx, elt in voter.items():

                    ensemble_proba += elt.predict_proba(X)

                    if verbose == 1:
                        pbar.update(idx)

                if verbose == 1:
                    pbar.update(n_iter)

                return ensemble_proba / n_iter

            # if weights is not None:
            for idx, elt in voter.items():

                ensemble_proba += weights[idx] * elt.predict_proba(X)

                if verbose == 1:
                    pbar.update(idx)

            if verbose == 1:
                pbar.update(n_iter)

            return ensemble_proba

        # end calculate_probas ----

        if self.n_jobs is None:

            if self.verbose == 1:
                pbar = Progbar(self.n_estimators)

            if weights is None:

                return calculate_probas(self.voter, verbose=self.verbose)

            # if weights is not None:
            self.weights = weights

            return calculate_probas(self.voter, weights, verbose=self.verbose)

        # if self.n_jobs is not None:
        def predict_estimator(m):
            try:
                return self.voter[m].predict_proba(X)
            except:
                pass

        if self.verbose == 1:

            preds = Parallel(n_jobs=self.n_jobs, prefer="threads")(
                delayed(predict_estimator)(m)
                for m in tqdm(range(self.n_estimators))
            )

        else:

            preds = Parallel(n_jobs=self.n_jobs, prefer="threads")(
                delayed(predict_estimator)(m) for m in range(self.n_estimators)
            )

        ensemble_proba = 0

        if weights is None:

            for i in range(self.n_estimators):

                ensemble_proba += preds[i]

            return ensemble_proba / self.n_estimators

        for i in range(self.n_estimators):

            ensemble_proba += weights[i] * preds[i]

        return ensemble_proba

    def score(self, X, y, weights=None, scoring=None, **kwargs):
        """ Score the model on test set features X and response y. 

        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features

            y: array-like, shape = [n_samples]
                Target values

            scoring: str
                must be in ('explained_variance', 'neg_mean_absolute_error', \
                            'neg_mean_squared_error', 'neg_mean_squared_log_error', \
                            'neg_median_absolute_error', 'r2')
            
            **kwargs: additional parameters to be passed to scoring functions
               
        Returns: 
        
            model scores: {array-like}

        """

        preds = self.predict(X, weights, **kwargs)

        if scoring is None:
            scoring = "accuracy"

        # check inputs
        assert scoring in (
            "accuracy",
            "average_precision",
            "brier_score_loss",
            "f1",
            "f1_micro",
            "f1_macro",
            "f1_weighted",
            "f1_samples",
            "neg_log_loss",
            "precision",
            "recall",
            "roc_auc",
        ), "'scoring' should be in ('accuracy', 'average_precision', \
                           'brier_score_loss', 'f1', 'f1_micro', \
                           'f1_macro', 'f1_weighted',  'f1_samples', \
                           'neg_log_loss', 'precision', 'recall', \
                           'roc_auc')"

        scoring_options = {
            "accuracy": skm2.accuracy_score,
            "average_precision": skm2.average_precision_score,
            "brier_score_loss": skm2.brier_score_loss,
            "f1": skm2.f1_score,
            "f1_micro": skm2.f1_score,
            "f1_macro": skm2.f1_score,
            "f1_weighted": skm2.f1_score,
            "f1_samples": skm2.f1_score,
            "neg_log_loss": skm2.log_loss,
            "precision": skm2.precision_score,
            "recall": skm2.recall_score,
            "roc_auc": skm2.roc_auc_score,
        }

        return scoring_options[scoring](y, preds, **kwargs)



# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
import sklearn.metrics as skm2
from ..base import Base
from ..utils import misc as mx
from ..utils import matrixops as mo
from ..utils import lmfuncs as lmf
from sklearn.base import RegressorMixin


class BayesianRVFL2Regressor(Base, RegressorMixin):
    """Bayesian Random Vector Functional Link Network regression with two priors
    
    Attributes: 
    
        n_hidden_features: int
            number of nodes in the hidden layer

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out 
            of the training

        direct_link: boolean
            indicates if the original features are included (True) in model''s fitting or not (False)

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0: no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant). 
            Currently available: standardization ('std') or MinMax scaling ('minmax')

        col_sample: float
            percentage of features randomly chosen for training  

        row_sample: float
            percentage of rows chosen for training, by stratified bootstrapping    

        seed: int 
            reproducibility seed for nodes_sim=='uniform'

        s1: float
            std. dev. of init. regression parameters in Bayesian Ridge Regression

        s2: float
            std. dev. of augmented regression parameters in Bayesian Ridge Regression  

        sigma: float
            std. dev. of residuals in Bayesian Ridge Regression

        beta: array-like
            regression''s fitted parameters 

        Sigma: array-like
            covariance of the distribution of fitted parameters

        GCV: float
            return_std: boolean
            
        backend: str
            "cpu" or "gpu" or "tpu"                

    References:       

        - [1] Moudiki, T. (2020). Quasi-randomized networks for regression and classification, with two shrinkage parameters. Available at: 
          https://www.researchgate.net/publication/339512391_Quasi-randomized_networks_for_regression_and_classification_with_two_shrinkage_parameters
               
       
    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=0,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
        s1=0.1,
        s2=0.1,
        sigma=0.05,
        beta=None,
        Sigma=None,
        GCV=None,
        return_std=True,
        backend="cpu"
    ):

        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
            backend=backend
        )

        self.s1 = s1
        self.s2 = s2
        self.sigma = sigma
        self.beta = beta
        self.Sigma = Sigma
        self.GCV = GCV
        self.return_std = return_std

    def fit(self, X, y, **kwargs):
        """Fit BayesianRVFL2Regressor to training data (X, y)
        
        Args:        
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features
            
            y: array-like, shape = [n_samples]
                Target values
        
            **kwargs: additional parameters to be passed to 
                    self.cook_training_set
               
        Returns: 

            self: object

        """

        centered_y, scaled_Z = self.cook_training_set(y=y, X=X, **kwargs)

        n, p = X.shape
        q = self.n_hidden_features

        if self.direct_link == True:

            r = p + self.n_clusters

            block11 = (self.s1 ** 2) * np.eye(r)
            block12 = np.zeros((r, q))
            block21 = np.zeros((q, r))
            block22 = (self.s2 ** 2) * np.eye(q)

            Sigma_prior = mo.rbind(
                x=mo.cbind(x=block11, y=block12, backend=self.backend),
                y=mo.cbind(x=block21, y=block22, backend=self.backend),
                backend=self.backend,
            )

        else:

            Sigma_prior = (self.s2 ** 2) * np.eye(q)

        fit_obj = lmf.beta_Sigma_hat_rvfl2(
            X=scaled_Z,
            y=centered_y,
            Sigma=Sigma_prior,
            sigma=self.sigma,
            fit_intercept=False,
            return_cov=self.return_std,
            backend=self.backend,
        )

        self.beta = fit_obj["beta_hat"]

        if self.return_std == True:
            self.Sigma = fit_obj["Sigma_hat"]

        self.GCV = fit_obj["GCV"]

        return self

    def predict(self, X, return_std=False, **kwargs):
        """Predict test data X.
        
        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features.

            return_std: {boolean}, standard dev. is returned or not
            
            **kwargs: additional parameters to be passed to 
                    self.cook_test_set
               
        Returns: 
        
            model predictions: {array-like}

        """

        if len(X.shape) == 1:  # one observation in the test set only
            n_features = X.shape[0]
            new_X = mo.rbind(
                x=X.reshape(1, n_features),
                y=np.ones(n_features).reshape(1, n_features),
                backend=self.backend,
            )

        self.return_std = return_std

        if self.return_std == False:

            if len(X.shape) == 1:

                return (
                    self.y_mean
                    + mo.safe_sparse_dot(
                        self.cook_test_set(new_X, **kwargs),
                        self.beta,
                        backend=self.backend,
                    )
                )[0]

            return self.y_mean + mo.safe_sparse_dot(
                self.cook_test_set(X, **kwargs), self.beta, 
                backend=self.backend
            )

        else:  # confidence interval required for preds?

            if len(X.shape) == 1:

                Z = self.cook_test_set(new_X, **kwargs)

                pred_obj = lmf.beta_Sigma_hat_rvfl2(
                    X_star=Z,
                    return_cov=self.return_std,
                    beta_hat_=self.beta,
                    Sigma_hat_=self.Sigma,
                    backend=self.backend,
                )

                return (
                    self.y_mean + pred_obj["preds"][0],
                    pred_obj["preds_std"][0],
                )

            Z = self.cook_test_set(X, **kwargs)

            pred_obj = lmf.beta_Sigma_hat_rvfl2(
                X_star=Z,
                return_cov=self.return_std,
                beta_hat_=self.beta,
                Sigma_hat_=self.Sigma,
                backend=self.backend,
            )

            return (self.y_mean + pred_obj["preds"], pred_obj["preds_std"])

    def score(self, X, y, scoring=None, **kwargs):
        """ Score the model on test set features X and response y. 

        Args:
        
            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number 
                of samples and n_features is the number of features

            y: array-like, shape = [n_samples]
                Target values

            scoring: str
                must be in ('explained_variance', 'neg_mean_absolute_error', \
                            'neg_mean_squared_error', 'neg_mean_squared_log_error', \
                            'neg_median_absolute_error', 'r2')
            
            **kwargs: additional parameters to be passed to scoring functions
                
        Returns: 
        
            model scores: {array-like}
            
        """

        preds = self.predict(X)

        if self.return_std:  # if there are std. devs in the predictions
            preds = preds[0]

        if mx.is_factor(y):  # classification

            if scoring is None:
                scoring = "accuracy"

            # check inputs
            assert scoring in (
                "accuracy",
                "average_precision",
                "brier_score_loss",
                "f1",
                "f1_micro",
                "f1_macro",
                "f1_weighted",
                "f1_samples",
                "neg_log_loss",
                "precision",
                "recall",
                "roc_auc",
            ), "'scoring' should be in ('accuracy', 'average_precision', \
                               'brier_score_loss', 'f1', 'f1_micro', \
                               'f1_macro', 'f1_weighted',  'f1_samples', \
                               'neg_log_loss', 'precision', 'recall', \
                               'roc_auc')"

            scoring_options = {
                "accuracy": skm2.accuracy_score,
                "average_precision": skm2.average_precision_score,
                "brier_score_loss": skm2.brier_score_loss,
                "f1": skm2.f1_score,
                "f1_micro": skm2.f1_score,
                "f1_macro": skm2.f1_score,
                "f1_weighted": skm2.f1_score,
                "f1_samples": skm2.f1_score,
                "neg_log_loss": skm2.log_loss,
                "precision": skm2.precision_score,
                "recall": skm2.recall_score,
                "roc_auc": skm2.roc_auc_score,
            }

        else:  # regression

            if scoring is None:
                scoring = "neg_mean_squared_error"

            # check inputs
            assert scoring in (
                "explained_variance",
                "neg_mean_absolute_error",
                "neg_mean_squared_error",
                "neg_mean_squared_log_error",
                "neg_median_absolute_error",
                "r2",
            ), "'scoring' should be in ('explained_variance', 'neg_mean_absolute_error', \
                               'neg_mean_squared_error', 'neg_mean_squared_log_error', \
                               'neg_median_absolute_error', 'r2')"

            scoring_options = {
                "explained_variance": skm2.explained_variance_score,
                "neg_mean_absolute_error": skm2.mean_absolute_error,
                "neg_mean_squared_error": skm2.mean_squared_error,
                "neg_mean_squared_log_error": skm2.mean_squared_log_error,
                "neg_median_absolute_error": skm2.median_absolute_error,
                "r2": skm2.r2_score,
            }

        return scoring_options[scoring](y, preds, **kwargs)



"""Generalized 'linear' models for quasi-randomized networks (nonlinear models!)"""

# Authors: Thierry Moudiki
#
# License: BSD 3 Clear


import numpy as np
from ..base import Base
from ..optimizers import Optimizer


class GLM(Base):
    """Generalized 'linear' models for quasi-randomized networks (nonlinear models!)
    
    Attributes:
       
        n_hidden_features: int
            number of nodes in the hidden layer

        lambda1: float
            regularization parameter for GLM coefficients on original features

        alpha1: float
            controls compromize between l1 and l2 norm of GLM coefficients on original features

        lambda2: float
            regularization parameter for GLM coefficients on nonlinear features

        alpha2: float
            controls compromize between l1 and l2 norm of GLM coefficients on nonlinear features

        activation_name: str
            activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'

        a: float
            hyperparameter for 'prelu' or 'elu' activation function

        nodes_sim: str
            type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
            'uniform'

        bias: boolean
            indicates if the hidden layer contains a bias term (True) or not 
            (False)

        dropout: float
            regularization parameter; (random) percentage of nodes dropped out 
            of the training

        direct_link: boolean
            indicates if the original predictors are included (True) in model's 
            fitting or not (False)

        n_clusters: int
            number of clusters for 'kmeans' or 'gmm' clustering (could be 0: 
                no clustering)

        cluster_encode: bool
            defines how the variable containing clusters is treated (default is one-hot)
            if `False`, then labels are used, without one-hot encoding

        type_clust: str
            type of clustering method: currently k-means ('kmeans') or Gaussian 
            Mixture Model ('gmm')

        type_scaling: a tuple of 3 strings
            scaling methods for inputs, hidden layer, and clustering respectively
            (and when relevant). 
            Currently available: standardization ('std') or MinMax scaling ('minmax')

        optimizer: object 
            optimizer, from class nnetsauce.utils.Optimizer

        seed: int 
            reproducibility seed for nodes_sim=='uniform' and clustering       

    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        lambda1=0.01,
        alpha1=0.5,
        lambda2=0.01,
        alpha2=0.5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        optimizer=Optimizer(),
        seed=123,
    ):

        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            seed=seed,
        )
        
        self.lambda1 = lambda1
        self.alpha1 = alpha1
        self.lambda2 = lambda2
        self.alpha2 = alpha2
        self.optimizer = optimizer
        self.beta = None
        
        
    def compute_XB(self, X, beta=None, row_index=None):        
        
        if beta is not None:            
            
            if (row_index is None):
                
                return np.dot(X, beta)
            
            return np.dot(X[row_index,:], beta)
        
        # self.beta is not None in this case       
        if row_index is None:
            
            return np.dot(X, self.beta)
        
        return np.dot(X[row_index,:], self.beta)


    def compute_XB2(self, X, beta=None, row_index=None):   

        def f00(X):
            return np.dot(X, self.beta)

        def f01(X):
            return np.dot(X[row_index,:], self.beta)

        def f11(X):
            return np.dot(X[row_index,:], beta)

        def f10(X):
            return np.dot(X, beta)             

        h_result = {'00': f00,
                    '01': f01,
                    '11': f11,
                    '10': f10} 

        result_code = str(0 if beta is None else 1) 
        result_code += str(0 if row_index is None else 1)                              
        
        return h_result[result_code](X)

    
    
    def penalty(self, beta1, beta2, lambda1, lambda2, alpha1, alpha2):
        
        res = lambda1*(0.5*(1 - alpha1)*np.sum(np.square(beta1)) + alpha1*np.sum(np.abs(beta1)))            
        res += lambda2*(0.5*(1 - alpha2)*np.sum(np.square(beta2)) + alpha2*np.sum(np.abs(beta2)))            
        
        return res
        
           
        
    def compute_penalty(self, group_index, beta):                                                          
                
        return self.penalty(beta1=beta[0:group_index], 
                            beta2=beta[group_index:len(beta)], 
                            lambda1=self.lambda1, lambda2=self.lambda2, 
                            alpha1=self.alpha1, alpha2=self.alpha2)